package AccordionDrawer;


/**
 * Axis-aligned bounding box class for label drawing. Stores
 * information in screen coordintes, checks for overlap with other
 * bounding boxes.
 * 
 * @author  Tamara Munzner
 */

public class LabelBox extends Object implements Comparable{

    protected int bl[] = new int[2];
    protected int tr[] = new int[2];
    protected int blbg[] = new int[2];
    protected int center[] = new int[2];
    protected int extent[] = new int[2];
    protected int tl[] = new int[2];
	protected int br[] = new int[2];
    protected int frameNum;
    String name;
    int  labelheight;
    double drawAngle;
    int drawVertical;

    public LabelBox(int bottomLeft[], int topRight[], int bottomLeftBG[], int computedFrame, String name) {
	this.name = name;
	bl[0] = bottomLeft[0];
	bl[1] = bottomLeft[1];
	tr[0] = topRight[0];
	tr[1] = topRight[1];
	blbg[0] = bottomLeftBG[0];
	blbg[1] = bottomLeftBG[1];
	frameNum = computedFrame;
	//System.out.println("bl0 " + bl[0] + " tr0 " + tr[0] + " bl1 " + bl[1] + " tr1 " + tr[1] + " frame " + frameNum);
	//extent[0] = Math.abs(tr[0] - bl[0]);
	//extent[1] = Math.abs(tr[1] - bl[1]);
	//center[0] = bl[0]+extent[0]/2;
	//center[1] = tr[1]+extent[1]/2;
	int temp = tr[0] - bl[0];
	extent[0] = (temp < 0) ? -temp : temp;
	temp = tr[1] - bl[1];
	extent[1] = (temp < 0) ? -temp : temp;
	center[0] = bl[0]+(extent[0]>>1);
	center[1] = tr[1]+(extent[1]>>1);
	drawAngle = 0.0;
	drawVertical = 0;
	//System.out.println("made ex0 " + extent[0] + " ex1 " + extent[1] + " cen0 " + center[0] + " cen1 " + center[1]);
    }

	public LabelBox(int bottomLeft[], int bottomRight[], int topLeft[], int topRight[], int bottomLeftBG[], int computedFrame, String name, int labelheight, double angle, int drawVert) {
	   this.name = name;
	   bl[0] = bottomLeft[0];
	   bl[1] = bottomLeft[1];
	   br[0] = bottomRight[0];
	   br[1] = bottomRight[1];
	   tl[0] = topLeft[0];
	   tl[1] = topLeft[1];
	   tr[0] = topRight[0];
	   tr[1] = topRight[1];	
	   blbg[0] = bottomLeftBG[0];
	   blbg[1] = bottomLeftBG[1];
	   frameNum = computedFrame;
	   this.name = name;
	   this.labelheight = labelheight;
	   drawAngle = angle;
	   int temp = tr[0] - tl[0];
	   extent[0] = (temp < 0) ? -temp : temp;
	   temp = tr[1] - bl[1];
	   extent[1] = (temp < 0) ? -temp : temp;
	   center[0] = ((bl[0]+tl[0])>>1)+(extent[0]>>1);
	   center[1] = tr[1]+(extent[1]>>1);	 
	   drawVertical = drawVert;
	   //System.out.println("made ex0 " + extent[0] + " ex1 " + extent[1] + " cen0 " + center[0] + " cen1 " + center[1]);
	   }

    /**
     * @author Tamara Munzner
     * 
     * Test for intersection between this and another box (assuming
     * both are axis-aligned). Use cached frame number to judge
     * whether other box was created this frame and should be counted,
     * or it's left over from previous frame and can be ignored.
     * 
     * @param otherBox Box to check for intersection against this one
     */
    public boolean intersectBoxes(LabelBox b) {
	int D[] = new int[2];
	int fabs, fsum;
	//if (drawAngle == 0.0) drawAngle = 2;
	//double x = (labelheight+5)/Math.tan(Math.toRadians(drawAngle));
	double x = (labelheight+4)/Math.tan(Math.toRadians(drawAngle));
	if (frameNum < b.frameNum) return false;
	// check y direction first - common case is lots of leaves at
	// identical x distances but different y distances.
	for (int xy = 1; xy >=0 ; xy--) {
	    // compute difference of box centers, D = C1-C0
	    D[xy] = b.center[xy] - center[xy];
	    //fabs = Math.abs(D[xy]);
	   //fsum = extent[xy]/2 + b.extent[xy]/2;
	    fabs = (D[xy] < 0) ? - D[xy] : D[xy];
		fsum = (extent[xy]>>1) + (b.extent[xy]>>1);         
         // fsum = (b.size(xy)+size(xy)) >> 1;
//	    System.out.println(xy + " d " + D[xy] + " fabs " + fabs + " fsum " + fsum + " ext " + extent[xy] + " bext " + b.extent[xy] + " cent " + center[xy] + " bcent " + b.center[xy] + " fr " + frameNum + " bfr " + b.frameNum + " extent[xy]>>1 " + (extent[xy]>>1) + " extent/2 " + (extent[xy]/2));
	    if(drawVertical == 0)
	    {
	      if ( fabs > fsum)
		   return false;
	    } else if(drawVertical ==1)
	    {
		   if( Math.abs(b.bl[0] - br[0]) > x&& fabs > fsum )
		   return false;
	    } else
	    {
	    	if(Math.abs(b.tl[0] - tr[0]) > x && fabs > fsum )
	    	return false;
	    }
	}
  
	return true;
    }

    /** implement Comparable interface - sort on y values */
    public int compareTo(Object o) { 
	if (center[1]-(extent[1]>>1) == ((LabelBox)o).center[1]-(((LabelBox)o).extent[1]>>1)) return 0;
	else if (center[1] < ((LabelBox)o).center[1]) return -1;
	else return 1;
    }

	public boolean compareBtoT(LabelBox target, int low, int high)
	{
		return this.bl[1] < target.tr[1] && !(low==high);
	}
	
	public boolean bottomBiggerThanTop(LabelBox target)
	{
		return bl[1] >= target.tr[1];
	}
	
	public int pos(int xy)
	{
		return blbg[xy];
	}
    
	public int topRightPos(int xy)
	{
		return tr[xy];
	}

	/**
	 * @return
	 */
	public String getName() {
		return name;
	}
	
	public int size(int xy)
	{
		if (xy == 0)
			return tr[xy]-blbg[xy];
		else 
		{
			if(drawVertical != 2)		
			return blbg[xy]-tr[xy]+1;
		   else
		    return br[xy]-blbg[xy]+1;
		}
	}

};

