/*
 * Created on Apr 20, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package AccordionDrawer;

/**
 * @author hilde
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */

// Maintain a list of gridcells and indices that represent some split lines
//  - a split line is a horizontal or vertical collection of splits, in some order, with min and max stuck positions
//  - on creation of a gridcell (one that is not a leaf), check for a splitline that's already created, else add the gridcell as the splitline
//  - return the index when checking new gridcells

import java.util.Hashtable;
import java.awt.Color;

import gl4java.*;

public class SplitLine {
	
	public static final double defaultMinStuckValue = 0.001f;
	public static final double defaultMaxStuckValue = 0.999f;
	
	// the accordion drawer that owns this splitline
	AccordionDrawer ad;
	
	// true if this is the horizontal splitline
	boolean horizontal;
	
	// number of elements
	int size;

	// some collection, let's try a Vector
	//  - set size of vector to number of splitlines possible (bottomsize - 1)
	//  - position of gridcell is index of splitline on level 0 grid (lowest level, which is complete, at most bottomsize)
	//  - calculate the index by following path from root (left is 0, right is +1, shift left on descent, similar to russian peasant method)
	//
	// maybe try a hashtable later?
	//  - hash key is index of splitline on level 0 grid (lowest level, which is complete, at most bottomsize)
	//  - hash value is the grid cell
	//  - ignore leaf cells (just don't call here when they get created)
	// verify: the reverse lookup (grid cell to index) is never needed in the old code, it's stored in gridcell
	//  - adding new cells dynamically wouldn't work well, of course
	// for later perhaps: a reference count table for each index => maybe remove the gridcell when not drawn for more dynamic action?
//	Vector splits; // Vector of GridCells
//	double[] splitValues; // each in range of [0,1]
//	double[] absoluteValues; // all monotonic increasing in range of [0,1]
//	int[] computedFrame; // frame number for each splitline
	SplitCell[] splitCells;
	final double SplitResetValue = 0.5f;
	int rootIndex;
	
	double minStuckValue, maxStuckValue;
	
	// set fullGrid to true for a drawer that we know is full 
	public SplitLine(AccordionDrawer ad, boolean horizontal, double minStuckValue, double maxStuckValue)
	{
		this.ad = ad;
		this.horizontal = horizontal;
		this.minStuckValue = minStuckValue;
		this.maxStuckValue = maxStuckValue;
		size = (horizontal ? ad.bottomSize[AccordionDrawer.X] : ad.bottomSize[AccordionDrawer.Y]) - 1;
//		splits = new Vector();
		splitCells = new SplitCell[size];
//		absoluteValues = new double[size];
//		computedFrame = new int[size];
		rootIndex = getThisRoot(-1, size);
		if (size > 0)
			initSplitValuesRecurse(-1, rootIndex, size, rootIndex);
		resetSplitValues();
	}
	
	public int getThisRoot(int min, int max)
	{
		return (min + max + 1) / 2;
	}
	
	public int getSplitRoot(int min, int split, int max, boolean left)
	{
		if (left)
			return (min + split + 1) / 2;
		else
			return (split + max + 1) / 2;
	}
	
	private void initSplitValuesRecurse(int min, int mid, int max, int parent)
	{
		splitCells[mid] = new SplitCell();
		splitCells[mid].max = max;
		splitCells[mid].min = min;
		splitCells[mid].parent = parent;
		if (min + 1 == max) return;
		
		// dynamic? set another stopping condition such as computedFrame == -1
		splitCells[mid].kid = new int[2];
		splitCells[mid].splitValue = (double)(mid - min) / (double)(max - min);
		splitCells[mid].computedFrame = -1; // recompute the position of absolute location for each split line
		splitCells[mid].absoluteValue = -1;
		splitCells[mid].kid[0] = getSplitRoot(min, mid, max, true);
		splitCells[mid].kid[1] = getSplitRoot(min, mid, max, false);
		if (splitCells[mid].kid[0] != -1 && splitCells[splitCells[mid].kid[0]] == null)
			initSplitValuesRecurse(min, splitCells[mid].kid[0], mid, mid);
		else
			splitCells[mid].kid[0] = -1;
		if (splitCells[mid].kid[1] != size && splitCells[splitCells[mid].kid[1]] == null)
			initSplitValuesRecurse(mid, splitCells[mid].kid[1], max, mid);
		else
			splitCells[mid].kid[1] = -1;
	}
	
	
	private void resetSplitValuesRecurse(int min, int mid, int max)
	{
		if (min + 1 == max) return;
		// dynamic? set another stopping condition such as computedFrame == -1
		splitCells[mid].splitValue = (double)(mid - min) / (double)(max - min);
		splitCells[mid].computedFrame = -1; // recompute the position of absolute location for each split line
		splitCells[mid].absoluteValue = -1;
		resetSplitValuesRecurse(min, splitCells[mid].kid[0], mid);
		resetSplitValuesRecurse(mid, splitCells[mid].kid[1], max);
	}
	
	public void resetSplitValues()
	{
		resetSplitValuesRecurse(-1, rootIndex, size);

		// may need to adjust position to stuck values
		// - done in computePlaceThisFrame
		// must call computePlaceThisFrame before drawing anything after a reset
	}

	// set the min/maxline for the cell
	// set the splitCell index for the cell
	// call this for each internal node LeafGridCell
	// - internal tree nodes call this since they know their min/max lines
	// cellgeom should be set for cell before calling this
	// cellSplitIndex is the index into the splitCells data structure
	public void addCell(GridCell cell, int minLine, int maxLine)
	{
		int xy = horizontal ? 0 : 1;
		cell.minLine[xy] = minLine;
		cell.maxLine[xy] = maxLine;
	}
	
	// set the min/maxline for the cell
	// call this for each LeafGridCell
	// cellSplitIndex is the index into the splitCells data structure, above the cell
	public void addCell(GridCell cell, int splitCellIndex)
	{
		int xy = horizontal ? 0 : 1;
		cell.minLine[xy] = splitCellIndex-1;
		cell.maxLine[xy] = splitCellIndex;
	}
	
	
	// use absolute values to return the split index at the screen position
	public int getSplitLineIndex(double screenPosition)
	{
		if (screenPosition < minStuckValue || screenPosition > maxStuckValue)
			return -2;
		int min = -1, max = size;
		int mid = getThisRoot(min, max);
		while (min + 1 != max)
		{
			if (screenPosition < getAbsoluteValue(mid))
			{
				max = mid;
				mid = getThisRoot(min, mid);
			}
			else
			{
				min = mid;
				mid = getThisRoot(mid, max);
			}
		}
		if (screenPosition > getAbsoluteValue(min) || screenPosition < getAbsoluteValue(max))
			return mid;
		System.out.println("Error in getSplitLineIndex searching for the value: " + screenPosition);
		return -2;
	}
	

	/**
	 * @param f
	 */
	public void setMaxStuckValue(double f) {
		maxStuckValue = f;
	}

	/**
	 * @param f
	 */
	public void setMinStuckValue(double f) {
		minStuckValue = f;
	}

	/**
	 * @return
	 */
	public double getMaxStuckValue() {
		return maxStuckValue;
	}

	/**
	 * @return
	 */
	public double getMinStuckValue() {
		return minStuckValue;
	}
	
	// return the index of the minline
	public int getMinLine(int cellIndex)
	{
		return splitCells[cellIndex].min;
	}

	public int getMaxLine(int cellIndex)
	{
		return splitCells[cellIndex].max;
	}
	
	public int getParentSplitIndex(int cellIndex)
	{
		return splitCells[cellIndex].parent;
	}
	
	public boolean isRoot(int cellIndex)
	{
		return cellIndex == rootIndex;
	}
	
	// return true if this split is to the left of its parent
	public boolean isSplitLeftChild(int cellIndex)
	{
		if (isRoot(cellIndex)) return false;
		return splitCells[splitCells[cellIndex].parent].kid[0] == cellIndex;
	}
	
	/**
	 * Computes min[], max[], and splitPos, the minimum, maximum and
	 * split point coordinates for this GridCell in absolute window
	 * coordinates. 
	 * 
	 * Absolute coordinates are computed hierarchically from the
	 * relative position of the splits. Values computed are minimum
	 * boundary, maximum boundary, split position, and area. All these
	 * are computed in both window and screen coordinates. All these
	 * fields should never be accessed directly, only through the
	 * "get" access functions, so that this computation takes place
	 * when needed.
	 * 
	 * This data is updated only if it is stale, it is cached for the
	 * duration of a frame. The current frame number (frameNum) is the
	 * determiner of staleness, recompute whenever it's incremented.
	 * Note that incrementing can happen based in internal
	 * computation, not just an actual frame boundary visible to the
	 * user.
	 *
	 * @author   Tamara Munzner
	 *   */

	// compute new value for absoluteValues[cellIndex]
	public void computePlaceThisFrame(int cellIndex)
	{ // cell index is a split index of a cell
		int frameNum = ad.getFrameNum();
		if (cellIndex == -1 || cellIndex == size)
		{
////				System.out.println("cellIndex is a stuck value, can't compute place with this cellIndex: " + cellIndex);
			return;
		}
		if (cellIndex > -1 && cellIndex < size &&
			frameNum > splitCells[cellIndex].computedFrame)
		{
			int xy = horizontal ? AccordionDrawer.X : AccordionDrawer.Y;
			boolean thisCellIsLeft;
	 		if (!isRoot(cellIndex))
	 		{
				int parent = getParentSplitIndex(cellIndex);
				computePlaceThisFrame(parent);
				if (isSplitLeftChild(cellIndex))
				{
					int minLine = getMinLine(parent);
					double parMinPosition = minLine == -1 ?
						minStuckValue :
						splitCells[minLine].absoluteValue; // absolute position of min line
					double parSplitPosition = size == 0 ?
						maxStuckValue :
						splitCells[parent].absoluteValue;
					double range = parSplitPosition - parMinPosition; // size of parent min split
					splitCells[cellIndex].absoluteValue = parMinPosition + range * splitCells[cellIndex].splitValue;
					if (splitCells[cellIndex].absoluteValue < minStuckValue || splitCells[cellIndex].absoluteValue > maxStuckValue)
						System.out.println("Bad absolute value: " + cellIndex + " " + splitCells[cellIndex].absoluteValue + " " + minStuckValue + " " + maxStuckValue);
				}
				else
				{ // right split child
					int maxLine = getMaxLine(parent);
					double parMaxPosition = maxLine == size ?
						maxStuckValue :
						splitCells[maxLine].absoluteValue; // absolute position of max line
					double parSplitPosition = size == 0 ?
						minStuckValue :
						splitCells[parent].absoluteValue;
					double range = parMaxPosition - parSplitPosition; // size of parent max split
					splitCells[cellIndex].absoluteValue = parSplitPosition + range * splitCells[cellIndex].splitValue;
					if (splitCells[cellIndex].absoluteValue < minStuckValue || splitCells[cellIndex].absoluteValue > maxStuckValue)
						System.out.println("Bad absolute value: " + cellIndex + " " + splitCells[cellIndex].absoluteValue + " " + minStuckValue + " " + maxStuckValue);
				}
			}
			else
			{ // root node, parent doesn't really exist, this cell is in position 0 so it is on the left of the imaginary parent
				double myMinPosition = minStuckValue;
				double myMaxPosition = maxStuckValue;
				double range = myMaxPosition - myMinPosition;
				splitCells[cellIndex].absoluteValue = myMinPosition + range * splitCells[cellIndex].splitValue;
				if (splitCells[cellIndex].absoluteValue < minStuckValue || splitCells[cellIndex].absoluteValue > maxStuckValue)
					System.out.println("Bad absolute value: " + cellIndex + " " + splitCells[cellIndex].absoluteValue + " " + minStuckValue + " " + maxStuckValue);
			}
			splitCells[cellIndex].computedFrame = frameNum;
		}
	}

	private String intArrayToString(int[] array)
	{
		String s = "[" + array[0] + ", ";
		for (int i = 1; i < array.length-1; i++)
			s += array[i] + ", ";
		s += array[array.length-1] + "]";
		return s;	
	}
	private String doubleArrayDiffsToString(double[] array, double[] array2)
	{
		String s = "[" + (array[0]-array2[0]) + ", ";
		for (int i = 1; i < array.length-1; i++)
			s += (array[i]-array2[i]) + ", ";
		s += (array[array.length-1]-array2[array.length-1]) + "]";
		return s;	
	}
	private String doubleArrayToString(double[] array)
	{
		String s = "[" + array[0] + ", ";
		for (int i = 1; i < array.length-1; i++)
			s += array[i] + ", ";
		s += array[array.length-1] + "]";
		return s;	
	}
	private String doubleArrayToRangeString(double[] array)
	{
		if (array.length == 2)
			return "["+ (array[1]-array[0]) + "]";
		else if (array.length < 2) return "[]";
		String s = "[" + (array[1]-array[0]) + ", ";
		for (int i = 1; i < array.length-2; i++)
			s += (array[i+1]-array[i]) + ", ";
		s += array[array.length-1]-array[array.length-2] + "]";
		return s;	
	}

	public String toString()
	{
		String s = "too slow for debugging[";// + absoluteValues[0] + ", ";
//		for (int i = 1; i < splitValues.length-1; i++)
//			s += absoluteValues[i] + ", ";
//		s += absoluteValues[splitValues.length-1] + "]";
		return s;
	}
	
	public void setAbsoluteValue(int index, double value)
	{
		if (index > -1 && index < size)
			splitCells[index].absoluteValue = value;
		else
			System.out.println("Illegal absolute value index: " + index);
	}
	
	public double getAbsoluteValue(int index)
	{
		if (index == -1)
			return minStuckValue;
		if (index == size)
			return maxStuckValue;
		computePlaceThisFrame(index);
		if (index > -1 && index < size)
			return splitCells[index].absoluteValue;
		System.out.println("Illegal absolute value index: " + index);
		return -1f;
	}
	
	// returns index position (in the int array) of the index that is smaller than or equal to the root index
	// min and max index into splitIndices
	private int getIndexForRangeSplit(int[] splitIndices, int rootIndex, int min, int max)
	{
		if (rootIndex == splitIndices[min] || rootIndex == splitIndices[max])
		{
			System.out.println("Error: root index ("+ rootIndex+") shouldn't equal splitIndices[min] ("+splitIndices[min]+
					") or splitIndices[max] ("+splitIndices[max]+") when finding a split range");
			
		}
		int mid = (min+max)/2;
		while (min+1 < max && splitIndices[mid] != rootIndex)
		{
			if (splitIndices[mid] < rootIndex)
				min = mid;
			else
				max = mid;
			mid = (min+max)/2;
		}
		mid = (min+max)/2;
		return mid;
	}
	
	// newSplitValues is normalized (all values are between min=0 and max=1 from min to max) prior to calling
	// newSplitValues[min and max] may therefore be assumed to be 0 and 1 respectively and should not be modified
	private void resizeRanges(int[] splitIndices, double[] startValues, double[] endValues, int numSteps,
		int currRootIndex, int min, int max, Hashtable newToMove)
	{
		
//		System.out.println("Resizing ranges: min:" + min + " max:"+ max +" root:"+ rootIndex+" "+ doubleArrayDiffsToString(endValues, startValues));
 		if (max <= min)
 		{
			System.out.println("min == max in resize ranges: min:" + min + " max:"+ max +" root:"+ currRootIndex);
			return;
 		}
		if (max - min <= 1)
		{
			return; // base case, no splitlines to move inside a region with no ranges being resized
		}
		
		int index = getIndexForRangeSplit(splitIndices, currRootIndex, min, max);
		int xy = horizontal ? AccordionDrawer.X : AccordionDrawer.Y;
		double rootStartSize = startValues[max] - startValues[min];
		double rootEndSize = endValues[max] - endValues[min];
//		for (int i = min; i < max - 1; i++)
//		{
//			if (startValues[i] > startValues[i+1] || endValues[i] > endValues[i+1])
//			{
//				System.out.println("An error in position " + i + " " + startValues[i] + " " + startValues[i+1] + " " + endValues[i] + " " + endValues[i+1]);
//				return;
//			}	
//		}
		if (rootStartSize < 0 || rootEndSize < 0)
		{
			System.out.println("this is probably a synchronization problem");
			return; // what else can we do?  this is probably a synchronization problem
		}
		if (splitIndices[index] == currRootIndex) // some split is on the current root index, the easy case
			// 
		{
			// adjust the root index splitline location
			double rootOldLocation = startValues[index];
			double rootNewLocation = endValues[index];
			
			SplitTransition st = new SplitTransition(this, currRootIndex,
			(endValues[index]-endValues[min])/(rootEndSize),numSteps);
//			ad.toMove.add(st);

			newToMove.put(st.getHashKey(), st);
//			System.out.println("Added to move queue: " + ad.toMove);			
			
			// left side
			resizeRanges(splitIndices, startValues, endValues, numSteps,
				getSplitRoot(splitIndices[min], splitIndices[index], splitIndices[max], true),
				min, index, newToMove);
			
			// right side
			resizeRanges(splitIndices, startValues, endValues, numSteps,
				getSplitRoot(splitIndices[min], splitIndices[index], splitIndices[max], false),
				index, max, newToMove);
		}
		else // root is inside a range, this should be the common case (10 steps)
		{
			// step 1: move split at root index to right location
			double rootOldLocation = getAbsoluteValue(currRootIndex); // relative to screen width
			double rootOldSplit = splitCells[currRootIndex].splitValue; // relative to root region (min, max)
			double rootOldRangeSize = startValues[index+1] - startValues[index]; // size relative to index,index+1
			double rootNewRangeSize = endValues[index+1] - endValues[index]; // size relative to index,index+1
			if (rootNewRangeSize > 1 || rootNewRangeSize < 0)
				System.out.println("bad new root range size");
			double rootRatio = rootNewRangeSize/rootOldRangeSize; // ratio, can move split in index,index+1 range
			
			// relative to split region (index, index+1)
			double rootNewSplitTemp = rootRatio * // moving factor in index,index+1 range 
				(rootOldLocation-startValues[index]) + // normalize to the initial range position, old location is also a start position, both relative to screen width
				endValues[index]; // normalize to final range position 
			double rootNewSplit = (rootNewSplitTemp - endValues[min]) / (rootEndSize); // relative to root region (min, max)
			if (rootNewSplit > 1 || rootNewSplit < 0)
				System.out.println("bad new root split");
			double rootNewLocation = rootNewSplit * rootEndSize + endValues[min]; // relative to absolute location
			if (rootNewLocation > 1 || rootNewLocation < 0)
				System.out.println("bad new root location");
			SplitTransition st = new SplitTransition(this, currRootIndex, rootNewSplit,numSteps);
//			ad.toMove.add(st);
			newToMove.put(st.getHashKey(), st);
//			System.out.println("Added to move queue: " + ad.toMove);
			
			// step 2: store split on the right side boundary of root
			double tempEnd = endValues[index+1];    // store old range info for temporary over-write with current root
			double tempStart = startValues[index+1];
			int tempIndex = splitIndices[index+1]; // (will write this back on return from recursion)
			
			// step 3: write root split info into right side boundary of root
			endValues[index+1] = rootNewLocation;
			startValues[index+1] = rootOldLocation;
			splitIndices[index+1] = currRootIndex;
			
			// step 4: recursion on left side (min is same, max is index+1)
			if (rootNewLocation > 1 || rootNewLocation < 0)
				System.out.println("bad new root location");
			resizeRanges(splitIndices, startValues, endValues, numSteps,
				getSplitRoot(splitIndices[min], splitIndices[index+1], splitIndices[max], true),
//				(splitIndices[min]+rootIndex+1)/2,
				min, index+1, newToMove);
			
			// step 5: put right side split back in from temp variables (undo step 2, overwrite step 3 data but is still in variables)
			endValues[index+1] = tempEnd;
			startValues[index+1] = tempStart;
			splitIndices[index+1] = tempIndex;
			
			// step 6: store split on the left side boundary of root (see step 2 for similar action)
			tempEnd = endValues[index];
			tempStart = startValues[index];
			tempIndex = splitIndices[index];
			
			// step 7: write root split info into left side boundary of root (see step 3 for similar action)
			endValues[index] = rootNewLocation;
			startValues[index] = rootOldLocation;
			splitIndices[index] = currRootIndex;
			
			// step 8: recursion on right side (min is index, max is same)
			if (rootNewLocation > 1 || rootNewLocation < 0)
				System.out.println("bad new root location");
			resizeRanges(splitIndices, startValues, endValues, numSteps,
				getSplitRoot(splitIndices[min], splitIndices[index], splitIndices[max], false),
//				(splitIndices[index]+splitIndices[max]+1)/2,
				index, max, newToMove);
			
			// step 9: put left side split back in from temp variables (see step 5 for similar action, this undoes step 6)
			endValues[index] = tempEnd;
			startValues[index] = tempStart;
			splitIndices[index] = tempIndex;
			
			// done!
		}
	}
	
	// true when the first range was to the right of the min stuck position
	static private boolean createMin = false;
	
	// true when the last range was to the left of the max stuck position
	static private boolean createMax = false;
	
	// create range of indices of split lines for ranges in group
	// this includes the min/max stuck position split lines (if not already in a group)
	private int[] createIndexRanges(AbstractRangeList group)
	{
		int[] rangesTemp = group.getSplitIndices(horizontal);
		if (group.size() == 1 && 
			rangesTemp[0] == -1 && 
			rangesTemp[rangesTemp.length-1] == size)
			return null; // one range that covers the entire screen space 
		int minOffset = 0;
		createMin = false;
		createMax = false;
		int rangeSize = rangesTemp.length;
		if (rangesTemp[0] != -1) // first range doesn't include the min stuck position
		{
			rangeSize++;
			createMin = true;
			minOffset = 1;
		}
		if (rangesTemp[rangesTemp.length-1] != size) // last range doesn't include the max stuck position
		{
			rangeSize++;
			createMax = true;
		}
		int[] ranges = new int[rangeSize];
		for (int i = 0; i < rangesTemp.length; i++)
		{
			ranges[i+minOffset] = rangesTemp[i]; 
		}
		if (createMin) ranges[0] = -1;
		if (createMax) ranges[ranges.length-1] = size;
		return ranges;
	}
	
	private double getTotalExtent(double[] extent)
	{
		double totalExtent = 0f;
		for (int i = 0; i < extent.length; i++)
		{
			totalExtent += extent[i];
		}
		return totalExtent;
	}
	
	public void resizeForest(AbstractRangeList group, int numSteps, Hashtable newToMove, double inflateIncr)
	{
		if (group.size() == 0) return; // no ranges
		int[] ranges = createIndexRanges(group); // sets createMin, createMax
		if (ranges == null) return; // range covers everything
		if (group.size() == 1 &&
			getAbsoluteValue(ranges[1+(createMin?1:0)]) -
				getAbsoluteValue(ranges[0+(createMin?1:0)]) +
				inflateIncr < ad.minContextInside)
			return;
		double stuckRangeSize = maxStuckValue - minStuckValue;
		
		double[] startValues = new double[ranges.length]; // where the splitlines start
		double[] endValues = new double[ranges.length]; // where the splitlines will go
		double[] extent = group.getSizesOfAllRanges(this); // initial size of all ranges
		double oldTotalExtent = getTotalExtent(extent);
//		double inflateIncr = ad.inflateIncr;
		int rangesInPeriphery = (createMin ? 0 : 1) + (createMax ? 0 : 1);
//		double foo = ad.minContextPeriphery * (2-rangesInPeriphery);
//		double bar = (ranges.length - 1) * ad.minContextInside;
		double minGrowSize = stuckRangeSize - // stuck size 
				ad.minContextPeriphery * (2-rangesInPeriphery) - // room for peripheries 
				(ranges.length - 1) * ad.minContextInside; // room between ranges
		if (oldTotalExtent >= minGrowSize) // new potential size of ranges
		{
			System.out.println("Too much squishing, not going to grow (maybe too many ranges? " + ranges.length + ")");
			return; // don't grow any more
		}
		double minShrinkSize = stuckRangeSize - // stuck size
		ad.minContextPeriphery * rangesInPeriphery - // room for peripheries (range items)
		ranges.length * ad.minContextInside; // room between ranges
		double oldTotalNonExtent = stuckRangeSize - oldTotalExtent;
		
		for (int i = 0; i < ranges.length; i++)
		{
			startValues[i] = getAbsoluteValue(ranges[i]);
		}
		endValues[0] = startValues[0]; // stuck
		endValues[ranges.length-1] = startValues[ranges.length-1]; // stuck
		
		int numRealRanges = extent.length;
		int numRealNonRanges = numRealRanges - 1; // size if min/max stuck lines were in groups 
		if (createMin) // min wasn't initially a range line
			numRealNonRanges++;
		if (createMax) // max wasn't initially a range line
			numRealNonRanges++;
		
		int startAt = 0;
		int endAt = ranges.length - 1;
		// after this, startAt will point to the min splitline index of the minimum range to grow
		//  and endAt will point to the max splitline index of the maximum range to grow
		// if either createMin or Max, those ranges will be resized and the inflateIncr will be adjusted accordingly
	
		{
			if (oldTotalExtent + inflateIncr > minGrowSize)
				inflateIncr = minGrowSize - oldTotalExtent;
//			System.out.println("inflateIncr: " + inflateIncr);
			// note: since inflateIncr is being used here it should not be altered after this point
			double newTotalExtent = oldTotalExtent + inflateIncr;
			double newTotalNonExtent = oldTotalNonExtent - inflateIncr;
			double totalExtentRatio = newTotalExtent/oldTotalExtent;
			double totalNonExtentRatio = newTotalNonExtent/oldTotalNonExtent;
			
			double firstRange = startValues[startAt+1]-startValues[startAt];
			double lastRange = startValues[endAt]-startValues[endAt-1];
			if (createMin)
			{
			// the periphery shouldn't be reduced if this is true:
				if (firstRange < ad.minContextPeriphery)
				{
					System.out.println("Area before first range might be squished too small");
					endValues[startAt+1] = startValues[startAt+1];
				}
				else
				{ // adjust the periphery
					endValues[startAt+1] = startValues[startAt] + firstRange*totalNonExtentRatio;
					if (endValues[startAt+1]-endValues[startAt] < ad.minContextPeriphery)
						// note: this loses some exactness with the resizing by not refactoring wrt min periphery
						endValues[startAt+1] = endValues[startAt] + (double)ad.minContextPeriphery;
				}
				startAt++;
			}
			if (createMax)
			{
				if (lastRange < ad.minContextPeriphery)
				{
				//	note: this loses some exactness with the resizing by not refactoring wrt min periphery
					System.out.println("Area after last range might be squished too small");
					endValues[endAt-1] = startValues[endAt-1];
				}
				else
				{
					endValues[endAt-1] = startValues[endAt] - lastRange*totalNonExtentRatio;
					if (endValues[endAt]-endValues[endAt-1] < ad.minContextPeriphery)
						// note: this loses some exactness with the resizing by not refactoring wrt min periphery
						endValues[endAt-1] = endValues[endAt] - (double)ad.minContextPeriphery;
				}
				endAt--;
			}
			for (int i = startAt; i <= endAt-2; i+=2)
			{ // ranges to grow start at i, end at i+1
				endValues[i+1] = startValues[i+1] - startValues[i];
				endValues[i+1] *= totalExtentRatio;
				endValues[i+1] += endValues[i]; // end of last range
			  // start of next range
				endValues[i+2] = startValues[i+2] - startValues[i+1];
				endValues[i+2] *= totalNonExtentRatio;
				endValues[i+2] += endValues[i+1];
				if (endValues[i+2] - endValues[i+1] < ad.minContextInside)
				{
					double halfAddBack = ((startValues[i+2]-startValues[i+1]) - ad.minContextInside)/2;
					if (halfAddBack > 0)
					{
						endValues[i+1] = startValues[i+1] + halfAddBack;
						endValues[i+2] = startValues[i+2] - halfAddBack;
					}
					else
					{
						System.out.println("Not moving, min context inside is too small: " + (endValues[i+1]-endValues[i]) + " " + (endValues[i+3]-endValues[i+2]));
						endValues[i+1] = startValues[i+1];
						endValues[i+2] = startValues[i+2];
						System.out.println("Real: " + (endValues[i+1]-endValues[i]) + " " + (endValues[i+3]-endValues[i+2]));
					}
				}
			}
			
//			if (totalExtent + inflateIncr > 1)
//				inflateIncr = 1 - numRealNonRanges
		}

		
		// now resize the calculated ranges
		int min = 0, max = ranges.length - 1;
//		System.out.println("START: " + doubleArrayToRangeString(startValues));
//		System.out.println("END: " + doubleArrayToRangeString(endValues));
//		System.out.println("DIFF: " + doubleArrayDiffsToString(startValues, endValues));
		resizeRanges(ranges, startValues, endValues, numSteps, rootIndex, min, max, newToMove);
		// do movement transactions here
			
	}
	
	public void moveLine(int dragIndex, int dragPixelEnd, int staticIndex, int numSteps, Hashtable newToMove)
	{
		if (dragIndex == staticIndex)
			return; // huh?
		int xy = horizontal ? AccordionDrawer.X : AccordionDrawer.Y;
		if (dragIndex == -1 || dragIndex == size) return; // split is one of the stuck lines
		int[] range = {-1, dragIndex, staticIndex, size};
		double dragEnd = (double)ad.s2w(dragPixelEnd, xy);
		// shouldn't be able to drag the box out of the periphery range from the borders
		if (dragEnd < minStuckValue + ad.minContextPeriphery ||
			dragEnd > maxStuckValue - ad.minContextPeriphery)
			return;
		double staticEnd = getAbsoluteValue(staticIndex);
		double dragStart = getAbsoluteValue(dragIndex);
		// shouldn't be able to drag the box too small, keep context inside
		if (Math.abs(dragEnd - dragStart) < ad.minContextInside)
			return;
		double[] startValue = {minStuckValue, dragStart, staticEnd, maxStuckValue};
		double[] endValue = {minStuckValue, dragEnd, staticEnd, maxStuckValue};
		if (dragIndex > staticIndex)
		{
			range[1] = staticIndex;
			endValue[1] = staticEnd;
			startValue[1] = staticEnd;
			range[2] = dragIndex;
			endValue[2] = dragEnd;
			startValue[2] = dragStart;
		}
		int min = 0, max = range.length - 1;
		while (range[min] == range[min+1]) min++; // at least ignore statics that are
		while (range[max] == range[max-1]) max--; // on a stuck line
		
		resizeRanges(range, startValue, endValue, numSteps, rootIndex, min, max, newToMove);
	}

	// given the pixel, return the index splitLine on the minimum side of the gridCell chosen
	public int getMinIndexForPixelValue(int pixelPosition)
	{
		int xy = horizontal ? AccordionDrawer.X : AccordionDrawer.Y;
		double windowLocation = (double)ad.s2w(pixelPosition, xy);
		int minIndex = -1, maxIndex = size;
		int currIndex = rootIndex;
		double currLocation = getAbsoluteValue(currIndex);
		while (currIndex != maxIndex)
		{
			if (windowLocation <= currLocation)
			{
				maxIndex = currIndex;
				currIndex = getSplitRoot(minIndex, currIndex, maxIndex, true); //(minIndex + currIndex + 1) / 2;
			}
			else
			{
				minIndex = currIndex;
				currIndex = getSplitRoot(minIndex, currIndex, maxIndex, false); //(minIndex + currIndex + 1) / 2;
//				currIndex = (maxIndex + currIndex + 1) / 2;
			}
			currLocation = getAbsoluteValue(currIndex);
		}
//		System.out.println(pixelPosition + " " + minIndex);
		return minIndex;
	}
	
	public int getRootIndex()
	{
		return rootIndex;
	}
	
	// true when max is larger than min, thus this range is big enough to draw
	// this will assume that smallest non-big enough is when abs[min] == abs[max] (i.e. the same pixel)
	private boolean rangeSizeBigEnough(int min, int max, double pixelSize)
	{
		double size = getAbsoluteValue(max) - getAbsoluteValue(min);
		return size >= pixelSize;
	}
	
	// min, mid, max are all indices into split vectors (min should start at -1, max at size or max gridcell index)
	// recursion is exclusive (i.e. left and right sides don't both need to draw mid,
	//   arbitrary: give mid to the left side)
	public void drawRange(int min, int max, double pixelSize)
	{
		computePlaceThisFrame(min);
		computePlaceThisFrame(max);
		if (min+1 == max || !rangeSizeBigEnough(min, max, pixelSize))
		{
			ad.drawRange(min, max);
			ad.drawDensity(min, max);			
		}
		else // recursion
		{
			int mid = getThisRoot(min, max);
			drawRange(min, mid, pixelSize);
			drawRange(mid, max, pixelSize);
		}
	}
	
	// start recursion for drawingRanges, this is just for testing, it's not progressive
	public void drawAllRanges()
	{
		int xy = horizontal ? AccordionDrawer.X : AccordionDrawer.Y;
		double pixelSize = ad.s2w(1.0, xy);
		System.out.println("Draw all ranges");
		drawRange(-1, size, pixelSize);
	}

	/**
	 * @return
	 */
	public boolean isHorizontal() {
		return horizontal;
	}
	
	public void drawBackground(Color boxColor) {
		int xy = horizontal ? AccordionDrawer.X : AccordionDrawer.Y;
		float thecol[] = new float[3];
		int level = (int) (Math.log(size) / Math.log(2));
		int numRanges = 1;
		int currHop = (size+1) / 2;
		int currLevel = level;
		
		final int drawDepth = 5;
		double min[] = new double[2];
		double max[] = new double[2];
		min[xy^1] = ad.splitLine[xy^1].minStuckValue;
		max[xy^1] = ad.splitLine[xy^1].maxStuckValue;
		while (level - currLevel > drawDepth && currHop >= 1) {
			if (ad.colorgrid || ad.drawBackground) {
				int gridDepth = ad.gridDepth[xy];
				thecol[0] = 1.0f;
				thecol[1] = 1.0f - ((float) level) / gridDepth;
				if (numRanges == 1)
					thecol[2] = 1.0f;
				else if (currHop > 1)
					thecol[2] = .5f;
				else
					thecol[2] = .0f;
			} else {
				if (null == boxColor)
					return;
				boxColor.getRGBColorComponents(thecol);
			}
			double zplane =
				ad.colorgrid ? - (level + 2.0) : ad.hiliteplane;
			GLFunc gl = (GLFunc) ad.getGraphicsStuff();
			gl.glColor3f(thecol[0], thecol[1], thecol[2]);
			gl.glBegin(GLEnum.GL_POLYGON);
			int currPosition = 0;
			for (int i = 0; i < numRanges; i++)
			{
				currPosition += currHop;
				min[xy] = max[xy] = getAbsoluteValue(currPosition);
				gl.glVertex3d(min[0], max[1], zplane);
				gl.glVertex3d(min[0], min[1], zplane);
				gl.glVertex3d(max[0], min[1], zplane);
				gl.glVertex3d(max[0], max[1], zplane);
			}
			gl.glEnd();
			currLevel--;
			currHop = (currHop + 1) / 2;
			numRanges *= 2;
		}
		//	System.out.println("bbox: xmin " + min[0] + " xmax " + max[0] + " ymin " + min[1] + " ymax " + max[1] + " z " + z);
		//	System.out.println(" R " + thecol[0] + " G " + thecol[1] + " B " + thecol[2]);
	}
	
	protected void drawKidsGrid() {
		int xy = horizontal ? AccordionDrawer.X : AccordionDrawer.Y;
		int level = (int) (Math.log(size) / Math.log(2));
		int numRanges = 1;
		int currHop = (size + 1) / 2;
		int currLevel = level;

		final int drawDepth = 5;
		double min[] = new double[2];
		double max[] = new double[2];
		min[xy ^ 1] = ad.splitLine[xy ^ 1].minStuckValue;
		max[xy ^ 1] = ad.splitLine[xy ^ 1].maxStuckValue;
		while (level - currLevel > drawDepth && currHop >= 1) {

			GLFunc gl = (GLFunc) ad.getGraphicsStuff();
			float thecol[] = new float[3];
			if (ad.biggrid) {
				thecol[0] = currLevel & 1;
				thecol[1] = (currLevel & 2) / 2;
				thecol[2] = (currLevel & 4) / 4;
				gl.glLineWidth((float) (currLevel + 1.0));
			} else {
				gl.glLineWidth((float) 1.0);
				thecol[0] = thecol[1] = thecol[2] = .55f;
			}
			gl.glColor3f(thecol[0], thecol[1], thecol[2]);
			gl.glBegin(GLEnum.GL_LINES);
			int currPosition = 0;
			for (int i = 0; i < numRanges; i++) {
				currPosition += currHop;
				min[xy] = max[xy] = getAbsoluteValue(currPosition);
				gl.glVertex3d(min[xy], min[xy ^ 1], ad.gridplane);
				gl.glVertex3d(max[xy], max[xy ^ 1], ad.gridplane);
			}
			gl.glEnd();
			currLevel--;
			currHop = (currHop + 1) / 2;
			numRanges *= 2;
		}

	}
	
	// for testing with small datasets
	public void drawTicks()
	{
		GLFunc gl = (GLFunc) ad.getGraphicsStuff();
		gl.glLineWidth(1f);
		if (horizontal)
		{ // draw vertical tick marks
			SplitLine Ysplit = ad.getSplitLine(AccordionDrawer.Y);
			double minStuckY = Ysplit.minStuckValue;
			gl.glBegin(GLEnum.GL_LINES);
			for (int i = 0; i < splitCells.length; i++)
			{
				double absol = getAbsoluteValue(i);
				gl.glVertex3d(absol, 0, ad.interactionplane);
				gl.glVertex3d(absol, minStuckY, ad.interactionplane);
				
			}
			gl.glEnd();
		}
		else
		{ // draw horizontal tick marks
			SplitLine Xsplit = ad.getSplitLine(AccordionDrawer.X);
			double minStuckX = Xsplit.minStuckValue;
			gl.glBegin(GLEnum.GL_LINES);
			for (int i = 0; i < splitCells.length; i++)
			{
				double absol = getAbsoluteValue(i);
				gl.glVertex3d(0, absol, ad.interactionplane);
				gl.glVertex3d(minStuckX, absol, ad.interactionplane);
				
			}
			gl.glEnd();
		}
	}

	/**
	 * @return
	 */
	public int getSize() {
		return size;
	}
}
