
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionSequenceDrawer;

import java.util.*;
import java.awt.*;
import java.io.*;
import gl4java.*;

import AccordionDrawer.*;
import BitmapFont.*;

/**
 * An abstract class representing a multi-scale rectangular grid on which 
 * a tree consisting of horizontal and vertical line segments is drawn
 *
 * Fills in the fields of the abstract class AccordionDrawer based on
 * information from the Tree class.
 * 
 * @author  Tamara Munzner, Li Zhang
 * @version 
 * @see     AccordionDrawer.AccordionDrawer
 * @see     TreeJuxtaposer.AccordionTreeDrawerFinal
 * @see     AccordionDrawer.Tree
 * @see     AccordionDrawer.TreeNode
 * @see     AccordionDrawer.TreeEdge
 * @see     AccordionDrawer.GridCell
 */

//public class AccordionTreeDrawer extends AccordionDrawer {
public abstract class AccordionSequenceDrawer extends AccordionDrawer {
	String fname;
	//protected double pointoffset[] = new double[2];
	protected int labeloffset[] = new int[2];
	protected int labelbuffer[] = new int[2];

	/**
	 * framebuffer pixels from last mouseover label drawn
	 */
	protected float pixels[] = new float[2];
	/**
	 * GL rasterpos of last mouseover label drawn
	 */
	protected int drawnlabelpos[] = new int[2];
	/** 
	 * width and height of last mouseover label drawn
	 */
	protected int drawnlabelsize[] = new int[2];
	boolean biggerThanScreen = false;
	
	//how the labels are drawn
	// dealing with variable font size
	public int minFontHeight = 14;
	public int maxFontHeight = 14;
	public int popupFontHeight = 14;
	
	public static boolean complementFlag = false;
	
	protected double forestShrinkFactor;
	protected double pickoffset[] = new double[2];
	protected int pickFuzz;
	static String fontFile = "/OlduvaiFont";
	static BitmapFontsFamily bff = new BitmapFontsFamily(fontFile);
	static boolean fontInitialized = false;

	public static boolean fullyQualified;

	// benchmarking 
	public static int countDrawnFrame;
	public static int countDrawnScene;
	public static int cumFrameTime;
	
	public int[] defaultFocusRange = new int[2];

	public ArrayList sequences;

	public AccordionSequenceDrawer(ArrayList seq, int w, int h) {
		super(w, h);
		expandleaves = 0;
//		BinaryTree binaryTree = (BinaryTree)splitTree;
		horiz = true; // draw sequences horizontally

		int namewidth;
		if (!fontInitialized) {
			initializeFont();
			fontInitialized = true;
		}
		sequences = seq;
		for (int i = 0; i < seq.size(); i++)
		{
			namewidth = bff.stringWidth(getSequence(i).getName(), getMaxFontHeight());
			if (namewidth > nameStuckVal) nameStuckVal = namewidth;
		}
		labeloffset[X] = 2;
		labeloffset[Y] = 4;
		labelbuffer[X] = 10;
		labelbuffer[Y] = 1;

		pickoffset[0] = .1f;
		pickoffset[1] = .1f;
		pickFuzz = 5; // in pixels
		forestShrinkFactor = -0.9f; // -0.3;
		
		
		initCells();
		splitLine[Y].setMinStuckValue(0.021f);
		if (sequences.size() == 1)
			splitLine[X].setMinStuckValue(0.001f);
		else
			splitLine[X].setMinStuckValue(0.3f);

		flashGeomOld = null;
		doBox = false;

		defaultFocusRange[0] = 0;
		defaultFocusRange[1] = 0;
	}

	public boolean drawSpecial()
	{
		drawFrame();
		return true;
	}
	
	private void drawFrame()
	{
		long now;
		doFrameMove();
		int xy = splitLine[X].isHorizontal() ? X : Y;
		double pixelSize = s2w(1.0, xy);
		now  = System.currentTimeMillis();
		while (now - dynamicStart < dynamicTime && ToDrawQ.size() > 0)
//		while (ToDrawQ.size() > 0)
		{
			RangeInSequence r = (RangeInSequence)ToDrawQ.get(0);
			splitLine[X].drawRange(r.min, r.max, pixelSize);
			ToDrawQ.remove(0);
			now  = System.currentTimeMillis();
		}
//		requestRedraw();
		
		drawPostFrame();
		gl.glFlush();
		
		//System.out.println(" bla " + keepMoving + " " + ToDrawQ.size());
		if( keepMoving)
		{
			//System.out.println(" bla " + keepMoving + " " + ToDrawQ.size());
			requestRedraw();
			drawPostFrame();
		}
		else if (ToDrawQ.size() > 0)
		{
			keepDrawing = true;
			continueFrame();
		}
		else 
		{
			// don't start another frame
			// keepDrawing = false;
			if (colorgrid || drawBackground) splitLine[X].drawBackground(null);
			drawPostScene();
		}
			
	
	}
	public abstract ArrayList getColorsForRange( int minO, int maxO, int seqMin, int seqMax);
	public void drawRange(int min, int max)
	{
		
		// for each sequence, draw cells between min and max
		Iterator iter = sequences.iterator();
		while (iter.hasNext())
		{
			Sequence currSeq = (Sequence)iter.next();
			SiteNode n = currSeq.getNodeInRange(min+1, max);
			if (n != null)
			{
				ArrayList colors = getColorsForRange(min+1, max, currSeq.key, currSeq.key);
				if (colors == null || colors.size() == 0) {
					//gl.glEnable(gl.GL_DEPTH_TEST);
					n.drawInCell(n.getSiteColor(), objplane);
					//gl.glDisable(gl.GL_DEPTH_TEST);
				} else { // special group, alpha channel is for plane level => guaranteed visibility 
				//System.out.println("subpixel marked "+c+" omin "+c.objmin+" omax "+c.objmax);
					n.getCell().setDrewMarkedAttachedFrame(getFrameNum());
					for (int i = 0; i < colors.size(); i++)
					{
						Color sampleColor = (Color)colors.get(i);
						//gl.glEnable(gl.GL_DEPTH_TEST);
						n.drawInCell(sampleColor, hiliteplane - .004*sampleColor.getAlpha());
						//gl.glDisable(gl.GL_DEPTH_TEST);
					}
				}
			}
			else
			{
//				System.out.println("Couldn't find a cell for sequence " + currSeq + " in this range: " + min + ", " + max);
			}
		}
	}
	
	// build all splitLines
	private void buildCells() {

		int numrows = sequences.size();
		int numcols = getSequence(0).getNumOfNodes();
		// numcols same for all seqs, padded.

		for (int b = 0; b < numrows; b++) { // horizontal binary tree init
			Sequence currSeq = ((Sequence) sequences.get(b));
			// go over each site node in the sequence
			for (int a = 0; a < numcols; a++) { // place all the leaves
				SiteNode n = currSeq.getNode(a);
				if (!n.isGap()) {
					GridCell c = new GridCell(this);
					c.setCellGeom(n);
					n.setCell(c);
					splitLine[X].addCell(c, a);
					splitLine[Y].addCell(c, b); // create vertical split for vertical binary tree
				}
			}
			
			// reverse lookup for gaps; this will be removed when all gaps are in a group
			GridCell lastGridCellSeen = null;
			for (int a = currSeq.getNumOfNodes()-1; a >=0; a--)
			{
				SiteNode n = currSeq.getNode(a);
				if (!n.isGap())
					lastGridCellSeen = n.getCell();
				else
					n.setCell(lastGridCellSeen); // for fast lookup of gap distance given first sitenode
			}
		}
	}



	/**
	 * Initialize variable font family for tree labels. fontFamily[i]
	 * is the font should be used when there enough space for font
	 * with size i+minFontSize. The reason this is stored in an array
	 * is to speed up the look up time. 
	 * 
	 * @author Li Zhang
	 * @see TreeNode.drawInCell
	 */
	private void initializeFont() {
		// bitmap fonts
		try {
			bff.read();
		} catch (IOException e) {
			System.out.println("cannot produce fonts");
			return;
		}
		System.out.println("bitmap fonts initialized" + " " + bff.nofonts);
	}

	/**
	 * Initialization done after graphics context has been established.
	 * @author Tamara Munzner
	 */
	public void init() {
		super.init();
		bff.setupGL(gl);
	}

	/**
	 * Initializes level 0 GridCells based on Tree tree, assign tree
	 * nodes to them. Set {@link #maxUsed maxUsed[]} in directions X
	 * and Y.
	 *
	 *
	 * @author   Tamara Munzner, Serdar Tasiran
	 * 
	 *
	 * @see      AccordionDrawer.GridCell
	 * @see      AccordionDrawer.Tree
	 * @see AccordionDrawer.TreeNode */

	public void initCells() {
	    
		int longestSoFar = 0;
		Iterator iter = sequences.iterator();
		while (iter.hasNext()) {
		Sequence s = (Sequence) iter.next();
		longestSoFar = (s.getNumOfNodes() > longestSoFar) ? s.getNumOfNodes() : longestSoFar;
		}
		bottomSize[0] = longestSoFar; 
		bottomSize[1] = sequences.size();
		initSplitLines();
		gridDepth = new int[2];
		double logBase2 = Math.log(2);
		for (int xy = 0; xy < 2; xy++)
		{
			gridDepth[xy] = (int)Math.ceil(Math.log(bottomSize[xy]) / logBase2);
			gridDepth[xy] = gridDepth[xy] <= 0 ? 1 : gridDepth[xy];
		}
		System.out.println("gridDepth: "+gridDepth[0] + " " +gridDepth[1] +" rows: bottomSize0 "+bottomSize[0]+" cols bottomSize1 "+bottomSize[1]);
		buildCells();
	}

	// x and y are pixel positions
	public CellGeom pickAttached(int x,	int y)
	{
		GridCell pickedCell = pickCell(x, y);
		if (pickedCell != null)
			return pickedCell.getCellGeom();
		return null;
	}
	
	// pick the sequence at pixel position y
	private Sequence pickSequence(int y)
	{
		double position = s2w(y, Y);
		int seqPosition = splitLine[Y].getSplitLineIndex(position);
		if (seqPosition < 0 || seqPosition > sequences.size() - 1)
			return null;
		return getSequence(seqPosition);	
	}
	
	// given some pixel position, return the gridCell
	public GridCell pickCell(int x, int y)
	{
		Sequence seq = pickSequence(y);
		if (seq == null) return null;
		SiteNode n = seq.pickSiteNode(s2w(x, X), splitLine[X]);
		if (n == null) return null;
		return n.cell;
	}

	public void addRangeToDrawQueue(RangeInSequence r)
	{
		if (r != null)
			ToDrawQ.add(r);
	}

	/** 
	 * Create box enclosing the subtree beneath input node/edge. The
	 * resulting box is the nearest GridLine boundaries to that point,
	 * which are cached so that redraws to erase the box can happen
	 * after the lines change position.
	 * @param cg TreeNode to act upon, passed in as CellGeom from
	 *  AccordionDrawer (where TreeNodes are not known)
	 * @see IntersectionBox
	 * @author Tamara Munzner */
	public InteractionBox makeBox(CellGeom cg) {
		if (null == cg)
			return null;
		SiteNode sn = null;
		if (cg instanceof SiteNode)
			sn = (SiteNode) cg;

		SiteNode minNode = sn;
		SiteNode maxNode = sn;
		GridCell minc[] = new GridCell[2];
		GridCell maxc[] = new GridCell[2];
		minc[X] = maxc[X] = minc[Y] = maxc[Y] = sn.getCell();
		int[] minSplit = {minc[X].getMinLine(X), minc[Y].getMinLine(Y)};
		int[] maxSplit = {maxc[X].getMaxLine(X), maxc[Y].getMaxLine(Y)};

		return new InteractionBox(minSplit, maxSplit, cg, this);
	}

	


	/**
	 * On mouseover, flash the object under the mouse and draw its
	 * label at maximum size.
	 *
	 * @author   Tamara Munzner
	 * 
	 *
	 * @see      AccordionDrawer.GridCell
	 * @see      AccordionDrawer.Tree
	 * @see      AccordionDrawer.TreeNode
	 */
	
	public void doFlash() {
		
		if (noflash
			|| (flashGeom == flashGeomOld && flashBox == flashBoxOld)
			|| null == glj
			|| !glj.gljMakeCurrent())
			return;
				
		doingFlash = true;
		SiteNode oldFlashSite=null, flashSite=null;
		ArrayList oldFlashColors = new ArrayList();
		Color oldFlashColor=null, flashColor=null;
		
		if (flashGeomOld instanceof SiteNode)
		{
			oldFlashSite = (SiteNode)flashGeomOld;
			oldFlashColor = oldFlashSite.getSiteColor();  // site color, needed when unflashing non-difference back to original color
			oldFlashColors.add(oldFlashColor);
		}
		if (flashGeom instanceof SiteNode)
		{
			flashSite = (SiteNode)flashGeom;
			flashColor = flashSite.getSiteColor();
		}
		
		Color objCol = (null == flashCol) ? flashColor : flashCol; // flashCol shouldn't be null...
				
		if (doBox && flashGeom != null) {
			flashBox = makeBox(flashGeom);
		}
		if (doBox && flashBoxOld != null){
			flashBoxOld.undraw();
		}
		
		

		boolean drawlabelsreal = drawlabels;
		//drawlabels = false;
		if (null != flashGeomOld && (flashGeomOld instanceof SiteNode && !((SiteNode)flashGeomOld).getLabel().equalsIgnoreCase(""))) {
		        int seqKey = ((SiteNode)flashGeomOld).getSequence().key;
		        int objKey = flashGeomOld.getKey();
//			ArrayList oldColors = getColorsForRange(objKey, objKey, seqKey, seqKey);
			ArrayList oldColors = getColorsForGridCell(flashGeomOld.getCell(), objKey);
			Color temp = (oldColors.size()) > 0 ? (Color)oldColors.get(0) : null;
			double oldPlane =
				(temp == null)
					? objplane
					: hiliteplane - .004f * temp.getAlpha();
			oldColors = (temp == null) ? oldFlashColors : oldColors;
			SiteNode siteNode = ((SiteNode) flashGeomOld);
			if (null != siteNode) 
			{
				double x= s2w(flashBoxOld.getMin(X)-2,X);
				double y;
				if(biggerThanScreen)
					y = s2w(flashBoxOld.getMax(Y),Y);
				else
					y= s2w(flashBoxOld.getMax(Y)+2,Y);
				
				// before you render the flash set the depth test
				// on LEQUAL because all geometry is in the depthmask
				// and when it flashes you want this geometry that
				// wins the test			
				//gl.glEnable(GLEnum.GL_DEPTH_TEST);
				//gl.glDepthFunc(gl.GL_LEQUAL);
				
				gl.glRasterPos3d(x,y, getLabelplane());
				gl.glDrawPixels( drawnlabelsize[0], drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, pixels);
				//System.out.println(" draw pixels later on" + x + " " + y + " " + biggerThanScreen);
				//gl.glDisable(GLEnum.GL_DEPTH_TEST);
				
				for (int i = 0; i < oldColors.size(); i++)
				{
					siteNode.drawInCell((Color)oldColors.get(i), oldPlane);
				}
				gl.glDisable(GLEnum.GL_DEPTH_TEST);
			}
		} else if (flashGeomOld instanceof SiteNode && ((SiteNode)flashGeomOld).isGap())
		{
			flashGeomOld.drawInCell(Sequence.gapColor, objplane);
		}
		if (null != flashGeom) {
			// we don't actually need newcol, we just need to know if
			// it's non-null to decide whether to paint the object in
			// the frontmost highlight plane or the rear object plane.
		        int seqKey = ((SiteNode)flashGeom).getSequence().key;
		        int objKey = flashGeom.getKey();
//			ArrayList newCols = getColorsForRange(objKey, objKey, seqKey, seqKey);
			ArrayList newCols = getColorsForGridCell(flashGeom.getCell(), objKey);
			Color c = null;
			if (newCols.size() > 0)
				c = (Color)newCols.get(0);
			double newPlane =
				(c == null)
					? objplane
					: hiliteplane - .004f * c.getAlpha();
			//System.out.println("plane " + objplane);
			SiteNode siteNode = ((SiteNode) flashGeom);
			if (null != siteNode) {
				
				// readpixels first 
				drawnlabelsize[0] = flashBox.getMax(X) - flashBox.getMin(X) + 4;
				if(  (flashBox.getMax(Y) + 4) > getWinsize(Y) ) 
				{
					biggerThanScreen = true;
					drawnlabelsize[1] = flashBox.getMax(Y) - flashBox.getMin(Y);	
				}
				else 
				{
					drawnlabelsize[1] = flashBox.getMax(Y) - flashBox.getMin(Y) + 4;
					biggerThanScreen = false;
				}
					
				drawnlabelpos[0] = flashBox.getMin(X) -2;
				if( biggerThanScreen )
					drawnlabelpos[1] = flashBox.getMax(Y);
				else
					drawnlabelpos[1] = flashBox.getMax(Y) +2;
				pixels = new float[drawnlabelsize[0]*drawnlabelsize[1]*3];
			
				
				// read pixels from frontbuffer
				gl.glReadBuffer(GLEnum.GL_FRONT);
				gl.glReadPixels( drawnlabelpos[0], getWinsize(Y) - drawnlabelpos[1], drawnlabelsize[0], drawnlabelsize[1], GLEnum.GL_RGB, GLEnum.GL_FLOAT, pixels);
				gl.glDrawBuffer(GLEnum.GL_FRONT);
				//gl.glEnable(GLEnum.GL_DEPTH_TEST);
				//gl.glDepthFunc(gl.GL_LEQUAL);
				siteNode.drawInCell(objCol, newPlane);
				//gl.glDisable(GLEnum.GL_DEPTH_TEST);
			}
		}
		if (doBox && flashBox != null)
		{
			flashBox.draw(rubberbandColor, flashBoxWidth, interactionplane);
		}
		
		if (doBox)
			flashBoxOld = flashBox;
		//System.out.println("ATD set fgo "+flashGeom);
		flashGeomOld = flashGeom;
		flashXOld = flashX;
		flashYOld = flashY;
		drawEnd();
		doingFlash = false;
		
	   
	}

	/** 
	 * Decide what GridCells to enqueue for drawing based on tree
	 * topology. We want to add parents and children of current
	 * CellGeom. We should find the GridCell corresponding to the
	 * higher-level TreeEdge associated with a TreeNode, not the
	 * level0 GridCell of the TreeNode itself.
	 * 
	 * @author Tamara Munzner
	 */

	public void drawNextFancy(CellGeom g) {
	    return;
	}

	protected void drawPreNewFrame() {
		countDrawnFrame = 0;
		countDrawnScene = 0;
		cumFrameTime = 0;
		drawNameLabels();
		//System.out.println("here");
		//drawSiteNodeDensity(gridDepth[0] > gridDepth[1] ? gridDepth[0] : gridDepth[1]);
		
	}
	
	private void drawLabelBox(String name, int fontheight, int[] cellMinPix, int[] cellMaxPix) {
		int rX = 0, X = 0;
		int rY = 1 ,Y = 1;
		if (!horiz) {
			X = 1;
			Y = 0;
		}
		// draw label only when namewidth < diff[X] AND fontheight < diff[Y] in vertical direction
		int diff[] = { cellMaxPix[X] - cellMinPix[X], cellMaxPix[Y] - cellMinPix[Y]};
		int namewidth = AccordionSequenceDrawer.bff.stringWidth(name, fontheight);
		if (namewidth > Sequence.maxFontLength) Sequence.maxFontLength = namewidth;
		GLFunc gl = (GLFunc) getGraphicsStuff();
		// multiply with the number of sequences - what a mess
		// SequenceJuxtaposer.sequenceCount is now public
		// it is all because we switched x and y, and now y is devided
		// by the number of sequences. unfortunately we use that as 
		// x. that causes trouble - 3 is a little offset
		if( fontheight <= (diff[Y] - 3)) {
			setColorGL(getLabelColor());
			gl.glRasterPos3d(
					 s2w(cellMinPix[X],X),//+(diff[X]/2.0)-(namewidth/2.0),X),
					 s2w(cellMinPix[Y]+(diff[Y]/2.0f)+(fontheight/2.0f),Y),
					 getLabelplane());
		bff.drawString(gl, name, fontheight);
		}
	}
	
	private void drawDensity(GridCell c, float depth) {
		float thecol[] = new float[3];
		float size = (float)c.getSize(X);
		float pixel = 1.0f/getWinsize(X);
		double levelsize = size;
		int level = 0;
		GridCell cell = c;
		float refpixel = (float) 10.0/getWinsize(X);
		float alpha = ( size - pixel ) / ( refpixel - pixel );
		float grayValue = (1 - alpha ) * 0.7f + alpha * 1.0f; 
		
		float diff = (c.objmax - c.objmin) / (float)getSequence(0).endNumOfNodes;

		
		//System.out.println("color " + grayValue + "counter " + depth);
		thecol[0] =  grayValue - diff;
		thecol[1] =  grayValue - diff;
		thecol[2] =  grayValue - diff;
		double zplane = colorgrid ?  - (5 + 2.0f) : hiliteplane;
		GLFunc gl = (GLFunc) getGraphicsStuff();
		gl.glColor3f(thecol[0], thecol[1], thecol[2]);
		gl.glBegin(GLEnum.GL_POLYGON);
		double[] min = {c.getMin(0),0.001f};
		double[] max = {c.getMax(0), splitLine[Y].getMinStuckValue()};
		gl.glVertex3d(min[0], max[1], zplane);
		gl.glVertex3d(min[0], min[1], zplane);
		gl.glVertex3d(max[0], min[1], zplane);
		gl.glVertex3d(max[0], max[1], zplane);
		gl.glEnd();
		//System.out.println("bbox: xmin " + min[0] + " xmax " + max[0] + " ymin " + min[1] + " ymax " + max[1]);
		//System.out.println(" R " + thecol[0] + " G " + thecol[1] + " B " + thecol[2] + " " + c.getSize(X,false) );
		}
	
	public void drawSiteNodeDensity(float depth)
	{
		for(int nodeCount = 0; nodeCount < ((Sequence)sequences.get(0)).getNumOfNodes(); nodeCount++)
		{
			// THIS IS EXPENSIVE!!!!!!!!!!
			GridCell cell = ((Sequence)sequences.get(0)).getNode(nodeCount).cell;
			
			double cellMin[] = {cell.getMin(0), 0.001f}; //cell.getMin(1, false)};
			double cellMax[] = {cell.getMax(0), splitLine[Y].getMinStuckValue()}; //cell.getMax(1, false)};
			int cellMinPix[] = {w2s(cellMin[0], 0), w2s(cellMin[1], 1)};
			int cellMaxPix[] = {w2s(cellMax[0], 0), w2s(cellMax[1], 1)};
			//drawLabelBox(name, fontHeight[1], cellMinPix, cellMaxPix);
			drawDensity(cell, depth);//drawDensity(size, )
		}

		
	}
	
	private void drawNameLabels()
	{
		if (sequences.size() == 1) { return; } 
		int[] fontHeight = {getMinFontHeight(), getMaxFontHeight()};
		int maxNameWidth = (int)(0.3 * getWinsize(X));
//		for (int fontSize = fontHeight[1]; fontSize >= fontHeight[0]; fontSize--)
//		{
			for (int seqCount = 0; seqCount < sequences.size(); seqCount++)
			{
				Sequence currSeq = getSequence(seqCount);
				String name = currSeq.getName();
				int nameWidth = bff.stringWidth(name, getMaxFontHeight());
				if (nameWidth > maxNameWidth) // need to cut off some characters from name
				{
//					System.out.println("trying to chop: " + name);
					int nameResidue = nameWidth - maxNameWidth; // residue to remove in pixels
					int charsToChop = (int)((double)nameResidue/(double)nameWidth * name.length());
//					System.out.println("residue: " + nameResidue + "  charsToChop: " + charsToChop);
					if (charsToChop <= name.length() && charsToChop >= 0)
						name = name.substring(0, name.length() - charsToChop);
					else
						name = "";
				}
				if (drawlabels && name.length() > 0) {
					double cellMin[] = {0.001f, currSeq.getNode(0).cell.getMin(Y)};
					double cellMax[] = {splitLine[X].getMinStuckValue(), currSeq.getNode(0).cell.getMax(Y)};
					int cellMinPix[] = {w2s(cellMin[0], 0), w2s(cellMin[1], 1)};
					int cellMaxPix[] = {w2s(cellMax[0], 0), w2s(cellMax[1], 1)};
					drawLabelBox(name, fontHeight[1], cellMinPix, cellMaxPix);
				}
			}
			
//		}
	}

	protected void drawPreContFrame() {
			countDrawnFrame = 0;
	}

	protected void drawPostScene() {
		drawPostFrame();
		// commented so now isn't needed, if needed, pass it properly, not a state machine!
//		if (dumpstats) {
//			stats.println(
//				" scene "
//					+ countDrawnScene
//					+ " frametimeCum "
//					+ cumFrameTime
//					+ " sceneTIME "
//					+ (now - continueStart));
//			//stats.flush();
//		}
	}

	protected void drawPostFrame() {
		long now = System.currentTimeMillis();
		countDrawnScene += countDrawnFrame;
		long frameTime = now - dynamicStart;
		cumFrameTime += frameTime;
		if (dumpstats)
			stats.print(" df " + countDrawnFrame + " frameTIME " + frameTime);
	}

	protected void drawBruteForce() {

	}

	public Sequence getSequence(int seqNum) {
		return (Sequence)sequences.get(seqNum);
	}
	public Canvas getCanvas() {
		return this;
	}
	public SiteNode getNodeForDrawing(int seqNum, int key) {
		return getSequence(seqNum).getNode(key);
	}

	public int getLabelBuffer(int xy) {
		return labelbuffer[xy];
	}
	public void setLabelBuffer(int buffer) {
		if (buffer < 1)
			buffer = 1;
		labelbuffer[X] = buffer;
		labelbuffer[Y] = buffer;
		requestRedraw();
	}
	public void setLabelBuffer(int buffer, int xy) {
		if (buffer < 1)
			buffer = 1;
		labelbuffer[xy] = buffer;
		requestRedraw();
	}
	public void increaseLabelBuffer(int xy) {
		setLabelBuffer(labelbuffer[xy] + 1, xy);
	}
	public void increaseLabelBuffer() {
		setLabelBuffer(labelbuffer[X] + 1);
	}
	public void decreaseLabelBuffer(int xy) {
		setLabelBuffer(labelbuffer[xy] - 1, xy);
	}
	public void decreaseLabelBuffer() {
		setLabelBuffer(labelbuffer[X] - 1);
	}

	public int getMaxFontHeight() {
		return maxFontHeight;
	}
	public void setMaxFontHeight(int fontheight) {
		if (fontheight < 1)
			fontheight = 1;
		maxFontHeight = fontheight;
		requestRedraw();
	}
	public void increaseMaxFontHeight() {
		setMaxFontHeight(maxFontHeight + 1);
	}
	public void decreaseMaxFontHeight() {
		setMaxFontHeight(maxFontHeight - 1);
	}

	public int getMinFontHeight() {
		return minFontHeight;
	}
	public void setMinFontHeight(int fontheight) {
		if (fontheight < 1)
			fontheight = 1;
		minFontHeight = fontheight;
		requestRedraw();
	}
	public void increaseMinFontHeight() {
		setMinFontHeight(minFontHeight + 1);
	}
	public void decreaseMinFontHeight() {
		setMinFontHeight(minFontHeight - 1);
	}
	public static int stringWidth(String name, int fontHeight)
	{
		return bff.stringWidth(name, fontHeight);
	}
	/* (non-Javadoc)
	 * @see AccordionDrawer.AccordionDrawer#subpixelDraw(AccordionDrawer.GridCell)
	 */
	public void subpixelDraw(GridCell c) {
		// TODO Auto-generated method stub
		
	}

	public void drawDensity( int minObj, int maxObj)
	{
		float thecol[] = new float[3];
		double zplane = colorgrid ?  - (5 + 2.0f) : hiliteplane;
		GLFunc gl = (GLFunc) getGraphicsStuff();
		
		double c1 = splitLine[X].getAbsoluteValue(minObj);
		double c2 = splitLine[X].getAbsoluteValue(maxObj);
		float size = (float)(c2 - c1) ; 
		float white_size = 5.0f/getWinsize(X);  
				
		thecol[0] = size / white_size;
		thecol[1] = size / white_size;
		thecol[2] = size / white_size;
		
		gl.glColor3f(thecol[0], thecol[1], thecol[2]);
		gl.glBegin(GLEnum.GL_POLYGON);
			double[] min = { c1, 0.001f };
			double[] max = { c2, splitLine[Y].getMinStuckValue()};
			gl.glVertex3d(min[0], max[1], zplane);
			gl.glVertex3d(min[0], min[1], zplane);
			gl.glVertex3d(max[0], min[1], zplane);
			gl.glVertex3d(max[0], max[1], zplane);
		gl.glEnd();
		
	}

};
