
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionSequenceDrawer;

/**
 * A class representing a two-dimensional orthogonal range tree.
 * RangeTree is a class that can answer range query in time (log n)^2. 
 * It implements the data structure of 2D range tree.
 *
 * Nodes of this tree are objects of type RangeNode.
 * 
 * @author  Yunhong Zhou
 * 
 * @see RangeNode
 * @see Tree2Tree
 * @see AccordionDrawer.Tree
 * @see AccordionDrawer.TreeNode
 *
 */


public class ColorTree 
{
    ColorNode root=null;
    int element=0;

    /**
     * Constructor: calls RangeTreeBuild to build the tree
     * @param dim dim = 0 represents a 2D range tree, dim = 1 represents a
     * 1D range tree. This notation is not intuitive. 
     * @param A an array of points
     */
    
    public ColorTree(RangeList list)
    {
    	//System.out.println("making a range tree with " + list);
    	RangeInSequence[] array = new RangeInSequence[list.getNumRanges()];
    	array = (RangeInSequence[])list.getRanges().toArray(array);
    	root = build1D(array, 0, array.length-1);
    }


    public boolean isEmpty(){ return root==null; }
	
    /** @return root */
    public ColorNode getRoot(){ return root; }
		
    /**
     * builds a 1D range tree using points from A[start] to A[end].
     * Assume the A is sorted by Y coordinate.
     * This will set the elemnt of the leaf(Y coordinate).
     *
     * @param A a list of points
     * @param start the start position from array A
     * @param end the end position from array A
     */
    ColorNode build1D(RangeInSequence[] A,int start,int end)    
    {	
		if( A.length == 0 )
			return null;	
	int mid=(start+end)/2;
	ColorNode tmp=new ColorNode(A[mid], null, null);

	if (end - start >= 2) // go left and right, this is the middle
	    tmp.setLeft(build1D(A,start,mid-1));
	if (end - start >= 1) // add to the right only (div rounds down, mid is left of range) 
		tmp.setRight(build1D(A, mid+1,end));
	return tmp;
    }
    
    
    /** 
     * Find a RangeInSequence
     */
    public RangeInSequence findSplitRangeNode(ColorNode v, RangeInSequence range){
    RangeInSequence curr = null;
	if(v!=null) {
		curr = v.getData();
		if (range.compareTo(curr) < 0) // this RIS is too big, go left
			return findSplitRangeNode(v.getLc(), range);
		if (range.compareTo(curr) > 0) // this RIS is too small, go right
			return findSplitRangeNode(v.getRc(), range);
		if (range.min > curr.max || range.max < curr.min)
			curr = null; // fix off by one errors
	}
	return curr; // this is null or the fallthrough from the conditional (a found range)
    }
    
    public String toString()
    {
    	return root.toString();
    }
    

};













