
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
	 notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
	 notice, this list of conditions and the following disclaimer in the
	 documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
	 nor the names of its contributors may be used to endorse or promote
	 products derived from this Software without specific prior written
	 permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package AccordionSequenceDrawer;
import AccordionDrawer.GridCell;

/**
 * A class representing a range (ordered pair (min,max)) of TreeNode
 * indices in a Tree.	  
 * 
 * @author  Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
 *
 */

public class RangeInSequence implements Comparable {
	int min;
	int max;
	Sequence seq;
	GridCell cell; // some cell in this range than we can use to get a site node color
	public RangeInSequence(int minimum, int maximum, Sequence seq) {
	min = minimum; 
	max = maximum;
	this.seq = seq;
	}
	public RangeInSequence intersects(RangeInSequence r, boolean draw)
{
	// if we're marking and it's the same sequence, intersection might happen
	// or if we're drawing, all intersections are valid, even across sequences
	if ((!draw && r.seq == seq) || draw) // intersect, sequence ownership doesn't matter here, should check sequence when determining color
	{
		if (r.max + 1 < min || max + 1 < r.min) // side by side is overlap, so add 1 (join adjacent groups)
			return null; // no intersection
		if (r.min <= min)
		{
			if (r.max >= max)
				return new RangeInSequence(r.min, r.max, seq); // r.min < min, r.max > max, this is in r
			else
				return new RangeInSequence(r.min, max, seq); // r.min < min, max > r.max, r starts first, this finishes
		}
		else // min < r.min
		{
			if (r.max >= max)
				return new RangeInSequence(min, r.max, seq); // r.max > max, this starts first, r finishes
			else
				return new RangeInSequence(min, max, seq); // max > r.max, r is in this
		}
	}
	return null; // diff asd and/or ranges don't intersect
}
	public int getMin() { return min; }
	public int getMax() { return max; }
	public int rangeLength() { return max - min + 1; }
	Sequence getSequence() { return seq; }
	public String toString()
	{
		return "(" +seq +":" + min + "->" + max + ")";
	}
	/**
	 * @return
	 */
	public Sequence getSeq() {
		return seq;
	}

	/**
	 * @param drawer
	 */
	public void setSeq(Sequence seq) {
		this.seq = seq;
	}
	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	 
	public int compareTo(Object o) {
		RangeInSequence other = (RangeInSequence)o;
		if (seq.key < other.seq.key) return -1; // this sequence is less
		if (seq.key > other.seq.key) return 1; // this sequence is greater
		// same sequence
		if (max + 1 < other.min) return -1; // this range is less, +1 since adjacent ranges should be combined
		if (min > other.max + 1) return 1; // this range is greater, +1 since adjacent ranges should be combined
		return 0; // some overlap, adjacent or identical ranges
	}
	
	public boolean equals(Object o) {
		return compareTo(o) == 0;
	}

	public double getSize(AccordionDrawer.SplitLine splitLine)
	{
		int minCellSplit = min-1;
		int maxCellSplit = max;
		return splitLine.getAbsoluteValue(maxCellSplit) -
			splitLine.getAbsoluteValue(minCellSplit); 
	}

};
