
/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
	 notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
	 notice, this list of conditions and the following disclaimer in the
	 documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
	 nor the names of its contributors may be used to endorse or promote
	 products derived from this Software without specific prior written
	 permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
*/

package AccordionSequenceDrawer;

import java.awt.Color;
import java.util.TreeSet;
import java.util.Iterator;

import AccordionDrawer.*;


/**
 * A class representing a list of RangeInTrees. A RangeInTree
 * represents ordered pair (min,max)) of TreeNodes.
 *
 * This class is a helper class for TreeJuxtaposer and keeps a
 * resizeable array of RangeInTree's.
 *
 * Note that each RangeInTrees item in a RangeList can be associated
 * with a different Tree.
 * 
 * @see TreeJuxtaposer.RangeInList
 * @see AccordionDrawer.TreeNode
 * 
 * @author  Tamara Munzner
 * */

public class RangeList extends AbstractRangeList {

	/**
	 * Resizeable array (linked list) of RangeInTree objects
	 */
//	  LinkedList ranges; // resizeable array of RangeInTree objects
	TreeSet ranges; // resizeable array of RangeInTree objects
	boolean enabled; 
	Color col;
	int key;
    
	/*
	 * Only highlight group items for this tree, not all the others
	 */ 
	boolean thisTreeOnly;
	ColorTree rangeTree;

	public RangeList(int i) {
	col = new Color((float).5, (float).5, (float).5);
	enabled = true;
	ranges = new TreeSet();
	thisTreeOnly = false;
	key = i;
	}
	public RangeList(int i, Color c, boolean on, boolean only) {
	key = i;
	col = c;
	enabled = on;
	ranges = new TreeSet();
	thisTreeOnly = only;
	}
    
	public void addRange(int min, int max, Sequence seq, boolean draw, boolean collideDetect) { // add range with intersection checking, this means the list is ordered
	RangeInSequence addSequence = new RangeInSequence(min, max, seq); // new range to add
	RangeInSequence intersectRange;
	if (collideDetect)
	{
		Iterator iter = ranges.iterator();
		while (iter.hasNext())
		{
			RangeInSequence currRange = (RangeInSequence)iter.next(); // currRange is a previously added range
			intersectRange = currRange.intersects(addSequence, draw); // find intersecting range or return null
			if (intersectRange != null)
			{
				iter.remove(); // found intersecting range, remove the overlapping range
				addSequence = intersectRange; // update the sequence being checked for intersections
			}
		}
	}
	insertInOrder(addSequence, draw);
}

	// this doesn't work
	public void removeRange(int min, int max, Sequence seq) {
		System.err.println("can't remove a range like this");
	RangeInSequence r = new RangeInSequence(min, max, seq);
	// delete if it it's identical to something previously passed in...
	ranges.remove(r);
	}

//	private void insertInOrder(RangeInSequence seq, boolean draw)
//	{
//	if (size() == 0 || !draw) // marking doesn't matter, just add the sequence
//	{
//		ranges.add(seq);
//		return;
//	}
//	if (seq.max < getRange(0).min)
//	{
//		ranges.add(0, seq);
//		return;
//	}
//	if (seq.min > getRange(size()-1).max)
//	{
//		ranges.add(seq);
//		return;
//	}
//	for (int i = 0; i < size() - 1; i++)
//	{
//		RangeInSequence curr = getRange(i);
//		RangeInSequence next = getRange(i+1);
//		if (seq.min > curr.max && seq.max < next.min)
//		{
//			ranges.add(i+1, seq);
//			return;
//		}
//	}
//	System.out.println("Error: couldn't place range " + seq + " in " + ranges);
//	}

	// treeset version
	private void insertInOrder(RangeInSequence seq, boolean draw)
	{
		ranges.add(seq);
	}

	public void clear() {
	ranges.clear();
	rangeTree = null;
	}

	void print() {
	RangeInSequence rt;
	try{
	RangeList temp = (RangeList)this.clone();
	for (int i = 0; i < ranges.size(); i++) {
		rt = (RangeInSequence)temp.ranges.first();
		temp.ranges.remove(rt);
		System.out.println(i + " treekey " + rt.seq.getKey() + " min " + rt.min + " max " + rt.max);
	}
	}
	catch (CloneNotSupportedException e)
	{
		System.out.println("Cloning should work for rangeLists");
		return;
	}

	}

	public boolean getEnabled() {return enabled;}
	public void setEnabled(boolean on) { enabled = on;}
	public void setColor(Color c) {col = new Color(c.getRed(), c.getGreen(), c.getBlue(), key);}
	public Color getColor() {if (col != null && col.getAlpha() != key) col = new Color(col.getRed(), col.getGreen(), col.getBlue(), key); return col;}
	public boolean getThisTreeOnly() { return thisTreeOnly;}
	void setThisTreeOnly(boolean on) { thisTreeOnly = on;}
	public int getKey() { return key; }
	public int size() { return ranges.size(); }
	public double[] getSizesOfAllRanges(SplitLine splitLine)
	{
		Iterator iter = ranges.iterator();
		double[] size = new double[size()];
		int i = 0;
		while (iter.hasNext())
		{
			RangeInSequence range = (RangeInSequence)iter.next();
			size[i++] = range.getSize(splitLine);
		}
		return size;
	}
	public String toString()
	{
		String returnString = key+"[";
		Iterator iter = ranges.iterator();
		RangeInSequence next = null;
		if (iter.hasNext())
			next =(RangeInSequence)iter.next(); 
		while (next != null && iter.hasNext())
		{
			returnString += next + ",\n\t";
			next = (RangeInSequence)iter.next();
		}
		if (ranges.size() > 0)
			returnString += next;
		returnString += "]";
		return returnString;
	}

	/**
	 * @return
	 */
	public TreeSet getRanges() {
		return ranges;
	}
	
	public RangeInSequence getHead() {
		return (RangeInSequence)ranges.first();
	}

// treeset can't support this function efficiently	
//	public RangeInSequence getRange(int i){
//		return (RangeInSequence)ranges.get(i);
//	}


	/**
	 * @param i
	 */
	public void setKey(int i) {
		key = i;
	}

	/**
	 * @param list
	 */
	public void setRanges(TreeSet list) {
		ranges = list;
	}
	
	public int getNumRanges()
	{
		return ranges.size();
	}

	/**
	 * @return
	 */
	public ColorTree getColorTree() {
		return rangeTree;
	}

	/**
	 * @param tree
	 */
	public void setColorTree(ColorTree tree) {
		rangeTree = tree;
	}
	/* (non-Javadoc)
	 * @see AccordionDrawer.RangeList#getSplitIndices(boolean)
	 */
	// return splitindices for each of the ranges, min and max, in order
	public int[] getSplitIndices(boolean horizontal) {
		int xy = horizontal ? AccordionSequenceDrawer.X : AccordionSequenceDrawer.Y;
		int[] returnArray = new int[ranges.size() * 2];
		Iterator iter = ranges.iterator();
		int i = 0;
		while (iter.hasNext())
		{
			RangeInSequence currRange = (RangeInSequence)iter.next(); 
			returnArray[i*2] = currRange.min-1;
			returnArray[i*2+1] = currRange.max;
			i++;
		}
		return returnArray;
	}
	/* (non-Javadoc)
	 * @see AccordionDrawer.AbstractRangeList#onlyThisAD(AccordionDrawer.AccordionDrawer)
	 */
	public AbstractRangeList onlyThisAD(AccordionDrawer d) {
		return this; // everything is in the same AD
	}

	// Ranges that call this must be in a single sequence only, i.e sequences with
	//  ranges that overlap in columns are not good to use here, only things that we
	//  use to grow can be used here, search results might have to be tweaked to
	//  have a draw group as well as the mark gorup
	public AbstractRangeList flipRangeToShrink(int xy, AccordionDrawer ad)
	{
		RangeList returnRangeList = new RangeList(this.key);
		int splitLineSize = ad.getSplitLine(xy).getSize()-1; 
		if (ranges.size() == 0)
		{
			return null;
		}
		Iterator flipIter = ranges.iterator();
		RangeInSequence prev = (RangeInSequence)flipIter.next();
		if (prev.min > 0)
			returnRangeList.addRange(0, prev.min-1, prev.seq, false, false);
		while( flipIter.hasNext())
		{
			RangeInSequence curr = (RangeInSequence)flipIter.next();
		if (prev.seq != curr.seq)
			System.out.println("Error, sequences don't match: " + curr.seq + " " + prev.seq);
		// do something
		returnRangeList.addRange( prev.max+1, curr.min-1, prev.seq, false, false );
		prev = curr; 
	}
	if (prev.max < splitLineSize)
		returnRangeList.addRange(prev.max+1, splitLineSize, prev.seq, false, false);
	return returnRangeList;
}


};

// at some point, consider whether we should deal with overlaps intelligently.
// what does this statement mean?  I know I wrote it, but it confuses me now (2004-06-22, james)
