
package SequenceJuxtaposer;

import java.util.*;
import java.awt.event.*;

import AccordionDrawer.*;
import AccordionSequenceDrawer.*;

/**
 * A class that fleshes out the abstract methods left in
 * AccordionTreeDrawer.
 *
 * In particular, mouse and keyboard events are mapped to
 * AccordionTreeDrawer methods.
 *
 * @author  Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
 * @version 
 * @see     AccordionDrawer.AccordionDrawer
 * @see     AccordionDrawer.AccordionTreeDrawer
 */

public class AccordionSequenceDrawerFinal extends AccordionSequenceDrawer {

	SequenceJuxtaposer sj;
	int dragStart[] = new int[2];
	int dragEnd[] = new int[2];
	int mousePrev[] = new int[2];
	int mouseNow[] = new int[2];
	SiteNode mouseOverNode;

	// actiongroup: what is supposed to grow when grow option (bigger/smaller) is pressed
	private int actiongroup;
	// markgroup: what is the selection colour that is currently being used
	//  (only applies to MARK[0-3], other ints are only associated with actiongroup)
	private int markgroup;

	// which direction to grow
	private int growDirection;
	final static int HORIZ = 0;
	final static int VERT = 1;
	final static int ALL = 2;
	private int growDirectionDefault = VERT;

	// does the selection (marking) of a node select just the node or the subtree
	private int selectionResolution;
	final static int NODE = 0;
	final static int SUBTREE = 1;
	private int selectionResolutionDefault = SUBTREE;

	private int keytarget;
	// developer mode for keytarget only
	final static int DEVELOPER = 11;
	// only let the keytarget and actionmode be set to NONE
	// any other state that can be seen externally should have real state (default value)
	final static int NONE = 12;
	private final static int keytargetDefault = SequenceJuxtaposer.flashGroup;

	private int actionmode;
	final static int MOUSEOVER = 0;
	final static int ST_FREEMOVE = 1;
	final static int ST_FREEMOVEAGAIN = 2;
	final static int ST_RESHAPE = 3;
	final static int RECT_CREATE = 4;
	final static int RECT_FREEMOVE = 5;
	final static int RECT_FREEMOVEAGAIN = 6;
	final static int RECT_RESHAPE = 7;
	final static int PICK_SEQUENCE = 8;
//	final static int PICK_SEQUENCE_DEV = 9; 
//	final static int LENS = 10;
	final static int MOVE_STUCKPOS = 11;
	
	private final static int actionmodeDefault = MOUSEOVER;

	private int hold = 0;
	

	boolean jump;
	SiteNode pickedSite;
	SiteNode pickedSiteOld;

	InteractionBox baseBox;
	RangeInSequence selectedRange; // range of selected nodes in a sequence
	InteractionBox sequenceBox; // an interaction box that selects a sequence

	/* State accessors and mutators */

	/* Action group */
	public void setActionGroup(int newActionGroup) {
		setActionGroupWrapper(newActionGroup, true);
	}
	public void setActionGroupWrapper(int newActionGroup, boolean repeat) {
		actiongroup = newActionGroup;
		if (repeat) 
			sj.asd.setActionGroupWrapper(newActionGroup, false);
		sj.observe();
	}
	public int getActionGroup() {
		return actiongroup;
	}

/*
 * 	final static int MOUSEOVER = 0;
	final static int ST_FREEMOVE = 1;
	final static int ST_FREEMOVEAGAIN = 2;
	final static int ST_RESHAPE = 3;
	final static int RECT_CREATE = 4;
	final static int RECT_FREEMOVE = 5;
	final static int RECT_FREEMOVEAGAIN = 6;
	final static int RECT_RESHAPE = 7;
	final static int PICK_SEQUENCE = 8;
	final static int MOVE_STUCKPOS = 11;
 */
	public String getStringForAction(int action)
	{
		switch (action)
		{
			case MOUSEOVER: return "MOUSEOVER";
			case ST_FREEMOVE: return "ST_FREEMOVE";
			case ST_FREEMOVEAGAIN: return "ST_FREEMOVEAGAIN";
			case ST_RESHAPE: return "ST_RESHAPE";
			case RECT_CREATE: return "RECT_CREATE";
			case RECT_FREEMOVE: return "RECT_FREEMOVE";
			case RECT_FREEMOVEAGAIN: return "RECT_FREEMOVEAGAIN";
			case RECT_RESHAPE: return "RECT_RESHAPE";
			case PICK_SEQUENCE: return "PICK_SEQUENCE";
			case MOVE_STUCKPOS: return "MOVE_STUCKPOS";
		}
		return "don't know what action " + action + " is";
	}

	/* Action mode */
	public void setActionMode(int newActionMode) {
		//System.out.println("setting action mode: " + newActionMode);
		actionmode = newActionMode;
//		sj.df.status.setText("set: " + getStringForAction(actionmode));
		sj.observe();
	}

	public int getActionMode() {
//		sj.df.status.setText("get: " + getStringForAction(actionmode));
		return actionmode;
	}

	public void setGrowDirection(int growDirection) {
		setGrowDirectionWrapper(growDirection, true);
	}
	public void setGrowDirectionWrapper(int growDirection, boolean repeat) {
		this.growDirection = growDirection;
		if (repeat)
			sj.asd.setGrowDirectionWrapper(growDirection, false);
		sj.observe();
	}
	public int getGrowDirection() {
		return growDirection;
	}

	public void setSelectionResolution(int selectionResolution) {
		setSelectionResolutionWrapper(selectionResolution, true);
	}
	public void setSelectionResolutionWrapper(
		int selectionResolution,
		boolean repeat) {
		this.selectionResolution = selectionResolution;
		if (repeat) 
			sj.asd.setSelectionResolutionWrapper(selectionResolution, false);
		sj.observe();
	}
	public int getSelectionResolution() {
		return selectionResolution;
	}

	public void setMarkGroup(int newMarkGroup) {
		setMarkGroupWrapper(newMarkGroup, true);
	}
	public void setMarkGroupWrapper(int newMarkGroup, boolean repeat) {
		markgroup = newMarkGroup;
		if (repeat)
			sj.asd.setMarkGroupWrapper(newMarkGroup, false);
		sj.observe();
	}
	public int getMarkGroup() {
		return markgroup;
	}

	AccordionSequenceDrawerFinal(
		ArrayList t,
		int w,
		int h,
		SequenceJuxtaposer thesj) {
		super(t, w, h);

		sj = thesj;
		sj.asd = this;
		setActionGroup(SequenceJuxtaposer.flashGroup);
		setActionMode(actionmodeDefault);
		keytarget = NONE;
		setGrowDirection(growDirectionDefault);
		setSelectionResolution(selectionResolutionDefault);
		jump = false;
		baseBox = null;
		changedMarks();
		
		disableAnnotationsByDefault();
	}

	// disable Annotation drawing by default and set groups to false
	void disableAnnotationsByDefault()
	{
		for( int i = 0; i<SequenceJuxtaposer.maxAnnotationNum; i++ )
		{
			sj.drawAnnotations( ((String)sj.foundFeatures.get(i)), false );	
		}
	}

	// implement abstract classes
	public ArrayList getColorsForRange(int objmin, int objmax, int seqMin, int seqMax) {
		return sj.getColorsForRange(objmin, objmax, seqMin, seqMax, (AccordionSequenceDrawer) this);
	}
	
	public ArrayList getColorsForGridCell(GridCell c, int objkey)
	{
		if (c.getCellGeom() == null)
			return null;
		SiteNode s = (SiteNode)c.getCellGeom();
		Sequence seq = s.getSequence();
		int seqKey = seq.getKey();
		return sj.getColorsForRange(objkey, objkey, seqKey, seqKey, (AccordionSequenceDrawer) this);
	}

//	/**
//	 * Make an InteractionBox from the GridCells nearest to the box of the mousedrag.
//	 * Expects dragStart[X]/Y, dragEnd[X]/Y to be set. 
//	 * 
//	 * @author Francois Guimbretiere, Tamara Munzner */
//
//	private void drawActiveSubtreeBox() {
//		// turn off original flashbox
//		InteractionBox temp = flashBox;
//		flashBox = null;
//		setFlash(null, rubberbandColor, -1, -1, false);// true);
//		CellGeom fg = null;
//		if (getActionGroup() == SequenceJuxtaposer.flashGroup) {
//			baseBox = makeBox(pickedSite);
//			if (null == baseBox) {
//				System.out.println("basebox null");
//				return;
//			}
//			flashBox = baseBox;
//			if (null != pickedSite)
//				fg = pickedSite;
//			else if (null != temp && null != temp.item)
//				fg = temp.item;
//		} else {
//			ArrayList forestRoots = sj.getGroupForest(getActionGroup(), this);
//			// array of TreeNodes
//			if (forestRoots.isEmpty()) {
//				actionmodeReset();
//				return;
//			}
//			SiteNode n = (SiteNode) (forestRoots.get(0));
//			baseBox = makeBox(n); // just use the first one, for now
//			if (null == baseBox) {
//				System.out.println("basebox2 null");
//				return;
//			}
//			fg = n;
//			flashBox = baseBox;
//		}
//		flashBoxWidth = 3f;
//		if (fg == null)
//		    fg = getSequence(0).getNode(0);
//		sj.setFocus(fg, this);
//		setFlash(fg, rubberbandColor, -1, -1, false); //true);
//	}

	private void actionmodeReset() {
		// reason: set action mode to mouseover when exiting other modes
		setActionMode(MOUSEOVER);

		baseBox = null;
		flashBox = null;
		pickedSite = null;
		pickedSiteOld = null;
		flashBoxWidth = 1f;
		sj.doFlashGeom(null, SequenceJuxtaposer.flashGroup, this, mouseNow[X], mouseNow[Y], false);
		keytarget = NONE; // squash keys in state machine
		sj.setQuasimode(false);
		mouseover(mouseNow[X], mouseNow[Y]);
	}

	/**
	 * Make an InteractionBox from the GridCells nearest to the box of the mousedrag.
	 * Expects dragStart[X]/Y, dragEnd[X]/Y to be set. 
	 * Restrict interactionBox to a sequence
	 * 
	 *  */

	private InteractionBox createSequenceBoxFromCells(int[] dragStart, int[] dragEnd) {
		// Convert from the start and end of a drag motion to absolute
		// min/max coordinates so that computation can be carried out
		// independent of the direction that the user chose to drag.
		int min[] = new int[2];
		int max[] = new int[2];
		
		Sequence startSeq;
		CellGeom start = pickAttached(dragStart[X], dragStart[Y]);
		if (start == null)
			return null;
		startSeq = ((SiteNode)start).getSequence();
		for (int xy = 0; xy < 2; xy++) {
			if (dragStart[xy] < dragEnd[xy]) {
				min[xy] = dragStart[xy];
				max[xy] = dragEnd[xy];
			} else {
				min[xy] = dragEnd[xy];
				max[xy] = dragStart[xy];
			}
		}
		
		int minPos, maxPos;
		SiteNode minNode, maxNode;
		CellGeom temp;
		
		temp = pickAttached(min[X], min[Y]);
		if (temp == null)
			return null;
		minPos = ((SiteNode)temp).getPosition();
		minNode = startSeq.getNode(minPos);
		
		temp = pickAttached(max[X], max[Y]);
		if (temp == null)
			return null;
		maxPos = ((SiteNode)temp).getPosition();
		maxNode = startSeq.getNode(maxPos);
		
		selectedRange = new RangeInSequence(minPos, maxPos, startSeq);
		
		int minCell[] = new int[2];
		int maxCell[] = new int[2];
		for (int xy = 0; xy < 2; xy++) {
			minCell[xy] = minNode.getCell().getMinLine(xy);
			maxCell[xy] = maxNode.getCell().getMaxLine(xy);
		}
		// TODO: make sure this still works, changed GridCell to int for minCell/maxCell
		int[] minSplit = {splitLine[X].getMinIndexForPixelValue(min[X]), splitLine[Y].getMinIndexForPixelValue(min[Y])};
		int[] maxSplit = {splitLine[X].getMinIndexForPixelValue(max[X])+1, splitLine[Y].getMinIndexForPixelValue(max[Y])+1};
		return new InteractionBox(minSplit, maxSplit, null, this);//)minCell, maxCell, null, this));

	}
	/*
		 * create lens from cell
		 * pick the cell min and max closest to your mouse position 
		 * create lenspixel buffer that store the 'old' actual tree pixels in
		 * the front buffer (see doLens() for questions why x and y lenspixels
		 * zoompixels is the actual lens
		 * 
		 * min
		 *   +--------+
		 *    |           |
		 *   +--------+ max
		 * 
		 * function is  calles ones in mousePressed()
		 * 
		 * @author Kristian Hildebrand	  
		 */
		private Lens createLensFromCells() {
//				int min[] = new int[2];
//				int max[] = new int[2];
//				int pos[] = new int[2];
//				int size[] = new int[2];
//			
//				// size of the lens is currently static here 
//				// TODO: change size in flowmenu or with a gui
//				size[X] = 200;
//				size[Y] = 100;
//			
//				max[0] = mouseNow[0] + size[X]/2;
//				min[1] =  mouseNow[1] - size[Y]/2;
//				min[0] = mouseNow[0] - size[X]/2;
//				max[1] =  mouseNow[1] + size[Y]/2;
//				pos[0]	= mouseNow[0];
//				pos[1]	=  mouseNow[1];		
//				
//				xlenspixels = new float[size[Y] * 3] [size[X] * 3];
//				ylenspixels = new float[size[X] * 3] [size[Y] * 3];
//				zoompixels = new float[size[X] * 3 * size[Y] * 3];
//			
//			
//			GridCell minCell[] = new GridCell[2];
//			GridCell maxCell[] = new GridCell[2];
//				for (int xy = 0; xy < 2; xy++) {
//					minCell[xy] = pickCellClosestTo(mouseNow[xy], xy, false);
//					maxCell[xy] = pickCellClosestTo(mouseNow[xy], xy, true);
//				}
//		
//			return (new Lens(minCell, maxCell, pos, size, null, this));
		return null;
		}
	

	/*
		 * function is called during mouseDragged()
		 * 
		 * current mouse positions are used to pick new cells
		 * store the splitpos of the picked minc and maxc
		 * in temp variables to switch back in original zoomed screen
		 * see doLens() for further information 
		 * 
		 * @author Kristian Hildebrand
		 */
		private void updateLens(int x, int y){
//			int pos[] = new int[2];
//			int tmp[] = new int[2];
//		
//			lens.pos[0]	= x;
//			lens.pos[1]	=   y;		
//		
//			lens.max[0] = mouseNow[0] + lens.size[X]/2;
//			lens.min[1] =  mouseNow[1] - lens.size[Y]/2;
//			lens.min[0] = mouseNow[0] - lens.size[X]/2;
//			lens.max[1] =  mouseNow[1] + lens.size[Y]/2;
//				
//	
//			for (int xy = 0; xy < 2; xy++) {
//				lens.minc[xy] = pickCellClosestTo(lens.pos[xy], xy, false);
//				lens.maxc[xy] = pickCellClosestTo(lens.pos[xy], xy, true);
//			}
//		
//			// splitpos is  in window coordinates between 0 and 1
//			lens.temp_min[Y] = lens.minc[Y].getSplitPos(Y,false);  
//			lens.temp_min[X] = lens.minc[X].getSplitPos(X,false);
//			lens.temp_max[Y] = lens.maxc[Y].getSplitPos(Y,false);
//			lens.temp_max[X] = lens.maxc[X].getSplitPos(X,false);
							
		}


	private int closeToStuck()
	{
		double mouse[] = {s2w(mouseNow[X], X), s2w(mouseNow[Y], Y)};
		double fuzz[] = {s2w(pickFuzz, Y), s2w(pickFuzz, Y)};
		for (int xy = X; xy <= Y; xy++)
		{
			if (Math.abs(mouse[xy] - splitLine[xy].getMinStuckValue()) < fuzz[xy])
			{
				//System.out.println("close to minstuck? " + xy);
				return xy + MIN_STUCK_X;
			}
			else if (Math.abs(mouse[xy] - splitLine[xy].getMaxStuckValue()) < fuzz[xy])
			{
				//System.out.println("close to maxstuck? " + xy);
				return xy + MAX_STUCK_X;
			}
		}
		return FAR_FROM_STUCK;
	}

	private void mouseover(int x, int y) {


		CellGeom pickedGeoms = pickAttached(x, y);
		if (null == pickedGeoms)
			return;
		else
			pickedSite = (SiteNode)pickedGeoms; 
		mouseOverNode = pickedSite;
		

		if (pickedSite == pickedSiteOld)
			return;

		double min[], max[];
		min = new double[2];
		max = new double[2];
		if (pickedSite != null) {

			min[X] = pickedSite.getMin(X);
			min[Y] = pickedSite.getMin(Y);
			max[X] = pickedSite.getMax(X);
			max[Y] = pickedSite.getMax(Y);
		}
		pickedSiteOld = pickedSite;
		sj.doFlashGeom(pickedSite, SequenceJuxtaposer.flashGroup, this, x, y, false);//true);
	}

	// use dragStart,End to reshape baseBox (if not null)
	// 
	private void reshaperectangle(boolean doOther) {

		if (null == baseBox) {
			actionmodeReset();
			return;
		}
		
		// CRFP: return 2 integers, pixel differences for baseBox movable lines in X and Y
		int[] rectBox = createRectFromPick(baseBox);
		if (rectBox == null)
			return;
			
		int numAnimSteps = 3;
		while (transitionAnimating)
			endAllTransitions(); // end transitions, new queue is going to be swapped in
		Hashtable newToMove = new Hashtable();
		for (int xy = X; xy <= Y; xy++)
		{
			splitLine[xy].moveLine(baseBox.moveLineIndex[xy], // the split to move
				rectBox[xy], // move it to the end position
				baseBox.stuckLineIndex[xy], // the split to stay
				numAnimSteps, newToMove);
			
		}
		toMove = newToMove; // swap in new queue
		incrementFrameNumber(); // so compute place this frame will update, this can be moved to after transitions are created
		requestRedraw(); // redraw the frame, this will enact the movements in the move queue
	}

	// what did this function do??
	private void remakerectangle() {
//		if (null == flashBox) {
//			actionmodeReset();
//			return;
//		}
//
//		// need to remake the baseBox with real cell data
//		// using boundaries of last rectmove flashBox
//		int[] dragStart = new int[2];
//		for (int xy = 0; xy < 2; xy++) {
//			// get just inside box
//			dragStart[xy] = flashBox.getMin(xy) + 1;
//			dragEnd[xy] = flashBox.getMax(xy) - 1;
//		}
//		CellGeom item = baseBox.item;
//		baseBox = createBoxFromCells(dragStart, dragEnd);
//		baseBox.item = item;
//
//		flashBox = baseBox;
//		setFlash(null, rubberbandColor, -1, -1, true);
//		// put back dragStart/End to real values
//		for (int xy = 0; xy < 2; xy++) {
//			baseBox.dragStart[xy] = mouseNow[xy];
//			dragEnd[xy] = mouseNow[xy];
//		}
	}

	private boolean mouseOutOfWindow(MouseEvent e)
	{
		return e.getX() <= 0 || e.getX() >= getWinMax(X) || e.getY() <= 0 || e.getY() >= getWinMax(Y);
	}

	public void mouseEntered(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
		displayMouseOver(e.getX(), e.getY());
		if (!sj.getQuasimode())
			requestFocus();
		else
			sj.wantsFocusInQuasi(this);
		doFlash();
	}
	public void mouseExited(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
		displayMouseOver(e.getX(), e.getY());
		System.out.println("left window" + sj.getQuasimode());
		if (!sj.getQuasimode())
			transferFocus();
		sj.clearGroup(SequenceJuxtaposer.flashGroup);
		doFlash();
	}

	// clicked call means there weren't any drag events.
	public void mouseClicked(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
		int action = getActionMode();
		int stuckValue = closeToStuck();
		if (MOUSEOVER == action && stuckValue != FAR_FROM_STUCK)
		{
//			System.out.println("picked stuck position: " + stuckValue);
			baseBox = createBoxFromCells(stuckValue);
			flashBox = baseBox;
			setFlash(null, rubberbandColor, mouseNow[X], mouseNow[Y], true);
			setActionMode(MOVE_STUCKPOS);
		}
		else if (RECT_FREEMOVE == action
			|| RECT_FREEMOVEAGAIN == action
			|| ST_FREEMOVE == action
			|| ST_FREEMOVEAGAIN == action) {
			// throw away return value, we just need the check of
			// whether faraway click means revert to mouseover
			if (null == baseBox)
				actionmodeReset();
		}
		mouseNow[X] = e.getX();
		mouseNow[Y] = e.getY();
	}

	public void mousePressed(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
			//System.out.println("mouse pressed");
		displayMouseOver(e.getX(), e.getY());
		mouseDragging = true;
		mouseNow[X] = e.getX();
		mouseNow[Y] = e.getY();
		if (pickedSite != null)
			defaultFocusRange[0] = pickedSite.getKey();
		
		
		for (int xy = 0; xy < 2; xy++) {
			dragStart[xy] = mouseNow[xy];
			dragEnd[xy] = mouseNow[xy];
		}
		if (baseBox != null)
			baseBox.setDragPoints(dragStart, dragEnd);

		if (MOUSEOVER == getActionMode()) {
			setActionMode(RECT_CREATE);
		}
		if (RECT_FREEMOVEAGAIN == getActionMode()) {
			setActionMode(RECT_RESHAPE);
			remakerectangle();
		} else if (RECT_FREEMOVE == getActionMode()) {
			
			if (null != flashBox) {
				setActionMode(RECT_RESHAPE);			
			} else {
				// no box yet. drag one out.
				flashBoxWidth = 3f;
				setActionMode(RECT_CREATE);
			}
		} else if (ST_FREEMOVE == getActionMode()) {
			setActionMode(ST_RESHAPE);
		} else if (ST_FREEMOVEAGAIN == getActionMode()) {
			setActionMode(ST_RESHAPE);
//		} else if (MouseEvent.BUTTON3 == e.getButton()) {
//			// go in Lens mode
//			setFlash(null, sj.backgroundColor, mouseNow[X], mouseNow[Y], true);
//				
//			setActionMode(LENS);
//			lens = createLensFromCells();
//			if( lens == null ) return;
//			// important for ignoring first frame
//			justPressed = true;
//			updateLens(mouseNow[X], mouseNow[Y]);
//			doLens();
//			lens.oldPos[X] = mouseNow[X];
//			lens.oldPos[Y] = mouseNow[Y];
//			System.out.println("lens on");
		}
	}

	public void mouseReleased(MouseEvent e) {
		if (RECT_RESHAPE == getActionMode())
		{
			baseBox.undraw();
			flashBox.undraw();
			doFlash();
			requestRedraw(); // too much drawing
			actionmodeReset();
//			System.out.println("release");
			return;
		
		}
//		if (mouseOutOfWindow(e))
//			return;
		displayMouseOver(e.getX(), e.getY());
		mouseDragging = false;
		
		CellGeom pickedGeo = pickAttached(e.getX(), e.getY());
				if( pickedGeo != null )
				{
					defaultFocusRange[1] = pickedGeo.getKey();
					if( defaultFocusRange[1] < defaultFocusRange[0])
					{
							int temp;
							temp = defaultFocusRange[0];
							defaultFocusRange[0] = defaultFocusRange[1];
							defaultFocusRange[1] = temp;
					}
					//System.out.println("focus range " + defaultFocusRange[0] + " " + defaultFocusRange[1]);
				}
				// if you don't pick a geometry, b/c it is a gap, you should set defaultFocusRAnge[1] set
				// to +1
				defaultFocusRange[1] = defaultFocusRange[0] + 1;
		
		
		if (RECT_CREATE == getActionMode()) {
			setActionMode(RECT_FREEMOVE);
		} else if (RECT_RESHAPE == getActionMode() || MOVE_STUCKPOS == getActionMode()) {
			/*
			 * release only, if <ctrl> is not hold
			 */
			if( hold == 0)
			{
				actionmodeReset();
				setActionMode(MOUSEOVER);
			}
			
		} else if (getActionMode() == ST_RESHAPE) {
			setActionMode(ST_FREEMOVEAGAIN);
		} else if (getActionMode() == PICK_SEQUENCE) {
			setActionMode(MOUSEOVER);
			if (selectedRange == null || selectedRange.getSeq() == null)
				return;
			sj.searchString = selectedRange.getSeq().getSubstring(selectedRange.getMin(), selectedRange.getMax()).toUpperCase();
			sj.searchSequences(SequenceJuxtaposer.foundGroupDraw, SequenceJuxtaposer.foundGroupMark);
			sj.requestRedrawAll();

//		} else if ( (getActionMode() == LENS) 	&& (MouseEvent.BUTTON3 == e.getButton())) {
//			// check if lens mode and third button is released then finish lens
//			// and go back in your originally state - much fun
//			System.out.println("lens off ");
//			if( lens == null ) return;
//			
//			updateLens(mouseNow[X], mouseNow[Y]);
//			doLens();
//			clearLens();
//			lens.oldPos[X] = mouseNow[X];
//			lens.oldPos[Y] = mouseNow[Y];		
//			justPressed = false;
//			setActionMode(MOUSEOVER);
		}
		
	}

	public void displayMouseOver(SiteNode n)
	{
		String seq = n.getSequence().getName();
		int min = seq.length() > 10 ? 10 : seq.length();
		sj.mouseOverView.setText(seq.substring(0, min) + ", " + (n.getKey()+1) + "(" + n.getName() + ")");
	}
	
	public void displayMouseOver(int x, int y)
	{
		CellGeom pickedGeoms = pickAttached(x, y);
			if (null == pickedGeoms)
				return;
		displayMouseOver((SiteNode) pickedGeoms);
	}

	public void mouseDragged(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
		displayMouseOver(e.getX(), e.getY());
		mouseNow[X] = e.getX();
		mouseNow[Y] = e.getY();
		int temp = mouseNow[X];
		if (!horiz) {
			int x = getWinMax(X);
			int y = getWinMax(Y);
			mouseNow[X] = mouseNow[Y] * x / y;
			mouseNow[Y] = temp * y / x;
		}
		dragEnd[X] = mouseNow[X];
		dragEnd[Y] = mouseNow[Y];
		if (baseBox != null)
			baseBox.dragEnd = dragEnd;
		
		if (PICK_SEQUENCE == getActionMode())
		{
			System.out.println(" PICK_SEQUENCE ");
			flashBoxWidth = 1f;
			sequenceBox = createSequenceBoxFromCells(dragStart, dragEnd);
			flashBox = sequenceBox;
			setFlash(null, rubberbandColor, mouseNow[X], mouseNow[Y], true);
		}
		else if (ST_FREEMOVE == getActionMode())
			setActionMode(ST_RESHAPE);
		else if (ST_RESHAPE == getActionMode()) {
			reshaperectangle(linkednav);
		}
		else if (getActionMode() == MOVE_STUCKPOS)
		{
//			System.out.println("moving stuck position");
			moveStuckPosition(baseBox);
		} else if ((RECT_CREATE == getActionMode())) {
			sj.setQuasimode(true);
			flashBox = null;
			setFlash(	null,
								rubberbandColor,
								mouseNow[X],
								mouseNow[Y],
								true);
			setActionMode(RECT_FREEMOVE);
			
			baseBox = createBoxFromCells(dragStart, dragEnd);
			flashBox = baseBox;
			setFlash(null, rubberbandColor, mouseNow[X], mouseNow[Y], true);
			
		} else if (RECT_RESHAPE == getActionMode()) {
			reshaperectangle(true);
		}

		// modification by jeffrey, state machine change for control key vs. drag start

		else if (RECT_FREEMOVE == getActionMode()) {
			flashBoxWidth = 3f;
			baseBox = createBoxFromCells(dragStart, dragEnd);
			flashBox = baseBox;
			setFlash(null, rubberbandColor, e.getX(), e.getY(), true);
//		} else if (LENS == getActionMode()) {
//			// when mouse is dragged it is obvious that the mouse pos is 
//			// changed
//			justPressed = false;
//			updateLens(mouseNow[X], mouseNow[Y]);
//			doLens();
//			lens.oldPos[X] = mouseNow[X];
//			lens.oldPos[Y] = mouseNow[Y];				
		}	
		mousePrev[X] = mouseNow[X];
		mousePrev[Y] = mouseNow[Y];
	}

	public void mouseMoved(MouseEvent e) {
		if (mouseOutOfWindow(e))
			return;
		displayMouseOver(e.getX(), e.getY());
		mouseNow[X] = e.getX();
		mouseNow[Y] = e.getY();
		int temp = mouseNow[X];
		if (MOUSEOVER == getActionMode())
			mouseover(mouseNow[X], mouseNow[Y]);
		mousePrev[X] = mouseNow[X];
		mousePrev[Y] = mouseNow[Y];
	}

	/** Used when hitting a key should trigger an atomic action.
	 * Alternative is a {@link keyPressed() keyPressed}/{@link
	 * keyReleased() keyReleased} pair that could be used in
	 * conjunction with mouseclicks.
	 * @author Tamara Munzner */
	public void keyTyped(KeyEvent e) {
		char key = e.getKeyChar();
		System.out.println("you typed " + key + " kt + " + keytarget);
		if(key == KeyEvent.VK_SPACE) {
					sj.setQuasimode(false);
					actionmodeReset();
					setActionMode(MOUSEOVER);			
		}
		if (key == 'f') {
			keytarget = SequenceJuxtaposer.flashGroup;
		} else if (key == 'd') {
			keytarget = SequenceJuxtaposer.diffGroupDraw;
		} else if (key == '1'){
			rendermode = true;
			reset();
		} else if (key == '2'){
			rendermode = true;
			reset();
		} else if (key == '3'){
			rendermode = true;
			reset();
		} else if (key == '0'){
			rendermode = false;
			reset();
		} else if (key == 'h') {
			keytarget = HORIZ;
			setGrowDirection(HORIZ);
		} else if (key == 'v') {
			keytarget = VERT;
			setGrowDirection(VERT);
		} else if (key == 'a') {
			keytarget = ALL;
			setGrowDirection(ALL);
			setSelectionResolution(SUBTREE);
		} else if (key == 'n') {
			keytarget = NODE;
			setSelectionResolution(NODE);
		} else if (key == '.' || key == '>') { // with or w/o shift
			keytarget = DEVELOPER;
		} else {
			if (keytarget != DEVELOPER) {
				if (key == 'g') {
					if (keytarget == SequenceJuxtaposer.siteGroup[0]
						|| keytarget == SequenceJuxtaposer.siteGroup[1]
						|| keytarget == SequenceJuxtaposer.siteGroup[2]
						|| keytarget == SequenceJuxtaposer.siteGroup[3]
						|| keytarget == SequenceJuxtaposer.siteGroup[4]
						|| keytarget == SequenceJuxtaposer.siteGroup[5]
						)
						setMarkGroup(keytarget);
					else {
						setMarkGroup(getMarkGroup() + 1);
						if (getMarkGroup() > SequenceJuxtaposer.siteGroup[5])
							setMarkGroup(SequenceJuxtaposer.siteGroup[0]);
					}
					sj.setGroupPriority(getMarkGroup());
				} else if (key == 'k') {
					sj.toggleLinkedNav();
				} else if (key == 'r') {
					sj.reset();

				} else if (key == 'R') {
					reset();
				} else if (key == 't') {
					if (keytarget == SequenceJuxtaposer.siteGroup[0]
						|| keytarget == SequenceJuxtaposer.siteGroup[1]
						|| keytarget == SequenceJuxtaposer.siteGroup[2]
						|| keytarget == SequenceJuxtaposer.siteGroup[3]
						|| keytarget == SequenceJuxtaposer.siteGroup[4]
						|| keytarget == SequenceJuxtaposer.siteGroup[5]
						|| keytarget == SequenceJuxtaposer.flashGroup
						|| keytarget == SequenceJuxtaposer.diffGroupDraw
						|| keytarget == SequenceJuxtaposer.gapGroup)
						setActionGroup(keytarget);
					else
						setActionGroup(keytargetDefault);
//					drawActiveSubtreeBox();
					setActionMode(ST_FREEMOVE);
					if (getActionGroup() == SequenceJuxtaposer.flashGroup)
						if (pickedSite != null)
							sj.setFocus(pickedSite, this);
						else {
							ArrayList forestRoots =
								sj.getGroupForest(getActionGroup(), this);
						// array of TreeNodes
							if (!forestRoots.isEmpty())
								sj.setFocus(
									((SiteNode) forestRoots.get(0)),
									this);
						}
				} else if (
					key == 'b' || key == 'B' || key == 's' || key == 'S') {

					boolean grow = (key == 'b' || key == 'B');
					int numSteps = (jump) ? 1 : getNumAnimSteps();
					boolean horiz = getGrowDirection() != VERT;
					boolean vert = getGrowDirection() != HORIZ;
					if (linkednav)
						sj.resizeGroup(
							getActionGroup(),
							grow,
							numSteps);
					else
						sj.resizeGroup(
							getActionGroup(),
							grow,
							numSteps);
					if (getActionGroup() == SequenceJuxtaposer.flashGroup)
						if (pickedSite != null)
							sj.setFocus(pickedSite, this);
						else {
							ArrayList forestRoots =
								sj.getGroupForest(getActionGroup(), this);
							// array of SiteNodes
							if (!forestRoots.isEmpty())
								sj.setFocus(
									((SiteNode) forestRoots.get(0)),
									this);
						}
					if (linkednav)
						sj.requestRedrawAll();
					else
						requestRedraw();
				} else if (key == 'm' || key == 'M') { // PICKSEQUENCE_DEV
//						System.out.println("pick_sequence_dev");
//						setActionMode(PICK_SEQUENCE_DEV);
						 

				}
				sj.requestRedrawAll();
			} else if (key == 'u') {
				sj.toggleShowDiffs();
			} else if (key == 'w') {
				sj.toggleDrawLabels();
			} else if (key == 'e') {
				actionmodeReset();
			} else {
				System.out.println("second tier");
				if (key == 'A') {
					sj.toggleShowAlg();
				} else if (key == 'N') {
					sj.requestRedrawAll();
				} else if (key == 'b') {
					sj.toggleBigGrid();
				} else if (key == 'g') {
					sj.toggleNoGrid();
				} else if (key == 'c') {
					sj.toggleColorGrid();
				} else if (key == 'T') {
					sj.toggleNoGeoms();
					requestRedraw();
				} else if (key == 'H') {
					sj.toggleNoFlash();
				} else if (key == 'o') {
					sj.toggleDrawOrder();
				} else if (key == 'k') {
					sj.toggleExpandLeaves();
				} else if (key == 'R') {
					sj.toggleLabelPosRight();
				} else if (key == 'w') {
					sj.increaseLineThickness();
				} else if (key == 'W') {
					sj.decreaseLineThickness();
				} else if (key == 'm') {
					sj.increaseLabelBuffer();
				} else if (key == 'M') {
					sj.decreaseLabelBuffer();
				} else if (key == 'x') {
					sj.increaseMaxFontHeight();
				} else if (key == 'X') {
					sj.decreaseMaxFontHeight();
				} else if (key == 'e') {
					sj.increaseMinFontHeight();
				} else if (key == 'E') {
					sj.decreaseMinFontHeight();
				} else if (key == 'B') {
					sj.toggleLabelDrawBack();
				} else if (key == 'p') {
					sj.toggleLabelPopup();
				} else if (key == 'D') {
					sj.toggleDimBrite();
				} else if (key == 'C') {
					sj.toggleDimColors();
				} else if (key == 'j') {
					jump = !jump;
				} else if (key == 's') {
					sj.toggleLabelTransp();
				} else if (key == 'O') {
					sj.toggleCacheRange();
				} else if (key == 'G') {
					sj.toggleGuarVis();
				} else if (key == 't') {
					sj.toggleCheckTime();
				} else if (key == 'J') {
					stats.flush();
					System.out.println("flushed stats\n");
				} else if (key == 'S') {
					sj.toggleDumpStats();
				} else if (key == 'F') {
					// the second display loop is so that the entire scene gets drawn, before going on to next iteration
					int i;
					System.out.println("brute force is disabled");
					setDumpStats(true);
					setCheckTime(true);
					setCacheRange(true);
					setDrawOrder(BRUTEFORCE);
					for (i = 0; i < 10; i++) {
						stats.println();
						stats.println();
						stats.print("brute checktime cache " + i + " ");
						display();
					}
					setCheckTime(false);
					for (i = 0; i < 10; i++) {
						stats.println();
						stats.println();
						stats.print("brute nochecktime cache " + i + " ");
						display();
					}
					setDrawOrder(FROMFOCUSCELL);
					for (i = 0; i < 10; i++) {
						stats.println();
						stats.println();
						stats.print("focus nochecktime cache " + i + " ");
						display();
					}
					setCheckTime(true);
					for (i = 0; i < 10; i++) {
						stats.println();
						stats.println();
						stats.print("focus checktime cache " + i + " ");
						display();
						while (keepDrawing)
							display();
					}

					stats.flush();
					stats.println();
					stats.println("MARK WHOLE TREE");
					stats.println();

					for (i = 0; i < 10; i++) {
						stats.println();
						stats.println();
						stats.print("mark checktime cache " + i + " ");
						display();
						while (keepDrawing)
							display();
					}
					setCheckTime(false);
					for (i = 0; i < 10; i++) {
						stats.println();
						stats.println();
						stats.print("mark nochecktime cache " + i + " ");
						display();
					}
					setCacheRange(false);
					for (i = 0; i < 10; i++) {
						stats.println();
						stats.println();
						stats.print("mark nochecktime nocache " + i + " ");
						display();
					}
					setCheckTime(true);
					for (i = 0; i < 10; i++) {
						stats.println();
						stats.println();
						stats.print("mark checktime nocache " + i + " ");
						display();
						while (keepDrawing)
							display();
					}
					stats.flush();
					stats.println();
					stats.println();
					stats.println("DONE WITH STATS");
					stats.flush();
					setDumpStats(false);
					System.out.println("done gathering stats");
				}
			}
			keytarget = NONE;
		}
		sj.observe();
	}

	private void newMouseOver(SiteNode originalNode, SiteNode newNode) {
				   sj.clearGroup(SequenceJuxtaposer.flashGroup);

				   String nodeName = newNode.getName();
				   displayMouseOver(newNode);
				   sj.addNodesToGroup(
						   newNode.getPosition(),
						   newNode.getPosition(),
						   SequenceJuxtaposer.flashGroup,
						   newNode.getSequence());
				   sj.doFlashGeom(
						   (CellGeom) newNode.cell.getCellGeom(),
						   SequenceJuxtaposer.flashGroup,
						   this,
						   getX(),
						   getY(),
						   true);
				   this.changedMarks();
				   this.pickedSite = newNode;
		   }



	// true if SiteNode cells (level 0 of quadtree) overlap horizontally
	private boolean inHorizRange(SiteNode a, SiteNode b) {
		final int A = 0, B = 1;
		double min[], max[];
		min = new double[2];
		max = new double[2];

		min[A] = a.cell.getMin(X);
		min[B] = b.cell.getMin(X);
		max[A] = a.cell.getMax(X);
		max[B] = b.cell.getMax(X);
		if (max[A] <= min[B] || max[B] <= min[A])
			return false;
		return true;
	}

	// find the child number of the originalNode as a child of the parent node
	private int findChildNumber(SiteNode parentNode, SiteNode originalNode) {
		return originalNode.getPosition();
	}

	// find the right aligned cell that is below (above) the original node, using the parent node as reference
	// - the last parameter (boolean) is used to determine whether to search down (true) or up (false)
	private SiteNode rightAlignFind(
		SiteNode parentNode,
		SiteNode originalNode,
		int originalNodeChildNumber,
		boolean goingDown) {
		return null;
	}

	//	//	mouseover parent
	//	// do nothing if currently no mouseover or at root node
	
		public void flashLeft()
		{
			if (sj.getGroup(SequenceJuxtaposer.flashGroup).getRanges().size() == 0)
					return;
			RangeInSequence rangeInSequence =
				(RangeInSequence) sj.getGroup(SequenceJuxtaposer.flashGroup).getHead();
			int firstNodeKey = rangeInSequence.getMin();
			SiteNode originalNode, childNode = null;
			originalNode = pickedSite; // childNode is originalNode asking the sequence for its neighbor
			if (originalNode == null || (childNode = originalNode.getSequence().getNode(--firstNodeKey)) == null)
					return;
			mouseOverNode = childNode; // keep the level for up/down traversal
		 	newMouseOver(originalNode, childNode);
			
		}
		
	//	//	mouseover first child 
	//	// do nothing if currently no mouseover or at terminal node
	
		public void flashRight()
		{
			if (sj.getGroup(SequenceJuxtaposer.flashGroup).getRanges().size() == 0)
				return;
			RangeInSequence rangeInSequence =
				(RangeInSequence) sj.getGroup(SequenceJuxtaposer.flashGroup).getHead();
			int firstNodeKey = rangeInSequence.getMin();
			SiteNode originalNode, childNode = null;
			originalNode = pickedSite; // childNode is originalNode asking the sequence for its neighbor
			if (originalNode == null || (childNode = originalNode.getSequence().getNode(++firstNodeKey)) == null)
				return;
			mouseOverNode = childNode; // keep the level for up/down traversal
			newMouseOver(originalNode, childNode);
		}
	//
	//
	//	//	mouseover sibling if at first child, cousin if at last child of common parent
	//	// do nothing if at bottommost node (bottommost has no tree edges below that share X coord)
	public void flashDown() {
		if (sj.getGroup(SequenceJuxtaposer.flashGroup).getRanges().size() == 0)
			return;
		RangeInSequence rangeInSequence =
			(RangeInSequence) sj.getGroup(SequenceJuxtaposer.flashGroup).getHead();
		int firstNodeKey = rangeInSequence.getMin();
		SiteNode originalNode, childNode = null;
		originalNode = pickedSite; // childNode is originalNode asking the sequence for its neighbor
		
		for(int i = 0; i < sequences.size()-1; i++){
				if ( ((Sequence)sequences.get(i)) == originalNode.getSequence())
				{
					// found the right one then pick the following
					if (originalNode == null || (childNode = ((Sequence)sequences.get(++i)).getNode(firstNodeKey)) == null)
							return;
					mouseOverNode = childNode; // keep the level for up/down traversal
					newMouseOver(originalNode, childNode);
				}
		}
		
	}
	
	//	//	mouseover sibling if at last child, cousin if at first child of common parent
	//	// do nothing if at topmost node
	public void flashUp() {
		if (sj.getGroup(SequenceJuxtaposer.flashGroup).getRanges().size() == 0)
				return;
			RangeInSequence rangeInSequence =
				(RangeInSequence) sj.getGroup(SequenceJuxtaposer.flashGroup).getHead();
			int firstNodeKey = rangeInSequence.getMin();
			SiteNode originalNode, childNode = null;
			originalNode = pickedSite; // childNode is originalNode asking the sequence for its neighbor
		
			for(int i = 1; i < sequences.size(); i++){
					if ( ((Sequence)sequences.get(i)) == originalNode.getSequence())
					{
						// found the right one then pick the following
						if (originalNode == null || (childNode = ((Sequence)sequences.get(--i)).getNode(firstNodeKey)) == null)
								return;
						mouseOverNode = childNode; // keep the level for up/down traversal
						newMouseOver(originalNode, childNode);	
						return;
					}
			}
		mouseOverNode = originalNode; // keep the level for up/down traversal
		//newMouseOver(originalNode, dNode);
		
	}

	public void keyPressed(KeyEvent e) {
		if (keytarget == DEVELOPER)
			return;
		if (sj.getQuasimode())
			return;
		char key = e.getKeyChar();
		int code = e.getKeyCode();

		if (code == KeyEvent.VK_LEFT)
			flashLeft();
		else if (code == KeyEvent.VK_RIGHT)
			flashRight();
		else if (code == KeyEvent.VK_DOWN)
			flashDown();
		else if (code == KeyEvent.VK_UP)
			flashUp();
		else
			//System.out.println("PRESS " + key + " code " + code);
			if (KeyEvent.VK_CONTROL == code || KeyEvent.VK_ALT == code) {
				sj.setQuasimode(true);
				flashBox = null;
				setFlash(
					null,
					rubberbandColor,
					mouseNow[X],
					mouseNow[Y],
					true);
				hold = 1;	
				setActionMode(RECT_FREEMOVE);
			} else if (KeyEvent.VK_SHIFT == code) {
				setActionMode(PICK_SEQUENCE);
			
			}
	}

	public void keyReleased(KeyEvent e) {
		if (keytarget == DEVELOPER)
			return;
		char key = e.getKeyChar();
		int code = e.getKeyCode();
		if (KeyEvent.VK_SHIFT == code
			|| KeyEvent.VK_CONTROL == code
			|| KeyEvent.VK_ALT == code) {
			sj.setQuasimode(false);
			actionmodeReset();
			hold = 0;
		}
	}

	public void seedQueue() {
//		System.out.println("Seeding queue for: " + this);
		sj.seedQueue(this,doDrawSpecial);
	}

	
};
