/*
   Copyright (c) 2002 Compaq Computer Corporation
   
   SOFTWARE RELEASE
   
   Permission is hereby granted, free of charge, to any person obtaining
   a copy of this software and associated documentation files (the
   "Software"), to deal in the Software without restriction, including
   without limitation the rights to use, copy, modify, merge, publish,
   distribute, sublicense, and/or sell copies of the Software, and to
   permit persons to whom the Software is furnished to do so, subject to
   the following conditions:
   
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   
   - Neither the names of Compaq Research, Compaq Computer Corporation
     nor the names of its contributors may be used to endorse or promote
     products derived from this Software without specific prior written
     permission.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
   IN NO EVENT SHALL COMPAQ COMPUTER CORPORATION BE LIABLE FOR ANY CLAIM,
   DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
   OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
   THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

package SequenceJuxtaposer;

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import java.io.*;

import AccordionDrawer.*;
import AccordionSequenceDrawer.*;
import SeqParser.*;

/**
 * A class representing a tree juxtaposer.
 * 
 * A TreeJuxtaposer takes a set of trees and build the data structures
 * for each pair for best corresponding nodes lookup and range
 * intersection queries.  The TreePairs class encapsulates all the
 * Tree2Tree classes required to relate pairs of trees in the list
 * "trees".
 * 
 *
 * @author Tamara Munzner, Serdar Tasiran, Li Zhang, Yunhong Zhou
 * @version 
 * @see     AccordionDrawer.Tree
 * @see     AccordionDrawer.TreeNode
 * @see     AccordionDrawer.CellGeom
 * @see     TreeJuxtaposer.TreePairs
 * @see     AccordionDrawer.AccordionTreeDrawer
 * @see     TreeJuxtaposer.AccordionTreeDrawerFinal
 */

public class SequenceJuxtaposer
	implements WindowListener, ActionListener, KeyListener {
		

	// this is important for the parser and sequences also
	static final public int MAX_STRING_LEN =  262144; // 2^18
	
	// Annotation stuff
	static public int maxAnnotationNum = 0;
	static public int seqNumWithMaxAnnotation = 0; 
	public Vector foundFeatures;
	
	Frame mainFrame;
	String title;

	//StateFrame stateFrame;
	//SettingsFrame settingsFrame;
	//IncrementalSearch searchFrame;
	//DebugFrame debugFrame;
	Panel drawPanel;
	GridLayout drawLayout;

	/**
	 * Resizeable array of AccordionTreeDrawerFinal objects
	 */
//	ArrayList treeDrawers;
	AccordionSequenceDrawerFinal asd;
	//    ArrayList ATDMatrix;

	/**
	 * Resizeable array of Tree objects
	 */
	int numSequences;
	ArrayList sequences;
	int maxSeqLength = 0;

	/**
	 * Rearrangable list of RangeLists (groups)
	 */
	LinkedList groups;
	Vector groupsByIndex;

	/**
	 * an array of TreePairs objects 
	 * each TreePair contains the necessary data structure for
	 * each pair
	 **/

	// ST: Don't understand this one well. Need help documenting.

//	SequencePairs TPs;

	Color backgroundColor;
	Color objectColor;
	Color labelColor;
	Color labelBackColor;
	Color labelHiColor;
	Color labelBackHiColor;
//	Color rubberbandColor;

	/** these numbers are permanent, but the ordering of the groups
	 * changes to reflect user interactions for which color wins out
	 * in a tie
	 */
	static int flashGroup; // flash group
	static int diffGroupDraw; // differences
	static int diffGroupMark; // differences
	static int siteGroup[] = new int[6]; // A, T, C, G, U, N
	static int gapGroup; // -
	static int foundGroupDraw;
	// use this group to draw the found group, ordered list of all sitenodes
	static int foundGroupMark;
	// use this group to mark the found group, unordered list of sequences of sitenodes
	
	//allocate memory, when you know, how many groups we have
	static int annotationGroupDraw[]; 
	static int annotationGroupMark[];
	static int rangeGapGroup;

	TextField searchField;
	Button searchButton;
	String searchButtonText = "Search";
	public Panel searchPanel;
	public String searchString;
	public final int searchStringCutoffSide = 15;
	// characters of text to display if string is too long
	public final int searchStringCutoffLength = 40;
	// max characters to display

	public TextField searchResultsView;
	public TextField mouseOverView;

	JPanel optionBar;

	CellGeom flashGeomOld;
	// for debug
	boolean showStructDiffs;
	boolean nMatchesEverything = true;
	boolean matrix;

	float linethickness;

	//static Status parser_status = new Status(System.in);

	boolean firstTime = true;
	private int sequenceArgs;

	boolean quasimode;
	AccordionSequenceDrawer wantsFocusATD;

	int edgeweightLevels;
	int edgeweightLevelUsed;
	
	float difference;

	static SequenceJuxtaposer self;

	UI ui;

	public DebugFrame df;

	public SequenceJuxtaposer() {
		
//		df = new DebugFrame();
//		df.show();

		self = this;
		AccordionSequenceDrawer.fullyQualified = false;
		// if true, all nodes will have fully qualified names
		// fully qualified names are good for trees that look like file systems
		// fully qualified names are bad for trees that look like phylogenies
		showStructDiffs = true;
		matrix = false;
		linethickness = 1.0f;
		// we use both swing and awt components,
		// so must disable the swing focus manager!
		javax.swing.FocusManager.disableSwingFocusManager();
		//doUI();
		
		/*controlPanel = ui.getControlPanel();
		searchFrame = ui.getSearchFrame();
		stateFrame = ui.getStateFrame();
		debugFrame = ui.getDebugFrame();
		settingsFrame = ui.getSettingsFrame();    
		*/

		//	ATDMatrix = new	ArrayList(); 
		sequences = new ArrayList();
		groups = new LinkedList();
		groupsByIndex = new Vector(7);

		backgroundColor = Color.getHSBColor(0.0f / 360f, 0f, 1f);
		// remember that rbcol is xor'ed against backcolor, 
		// so set to opposite of desired color...
//		rubberbandColor = Color.getHSBColor(0.0f / 360f, .0f, 0.3f);
		objectColor = Color.getHSBColor(0.0f / 360f, .0f, 0.15f);
		//labelColor = Color.getHSBColor(0f/360f,.0f,0.15f);
		//labelBackColor = Color.getHSBColor(0.0f/360f,0.0f,0.90f);
		//labelColor = Color.getHSBColor(100f/360f,1f,0.7f);
		labelColor = Color.getHSBColor(0f / 360f, 0f, 0f);
		labelBackColor = Color.getHSBColor(0.0f / 360f, 0.0f, 1f);
		labelHiColor = Color.getHSBColor(0.0f / 360f, .0f, 0.15f);
		labelBackHiColor = Color.getHSBColor(36f / 360f, 1f, 1f);

		Color flashCol = Color.getHSBColor(36f / 360f, 1f, 1f); //orange
		Color foundCol = Sequence.foundColor;
		// Color.getHSBColor(310f/360f,1f,1f); //pink
		Color diffCol = Sequence.diffColor;
		// Color.getHSBColor(0f/360f,1f,1f); //red
		//Color diffCol = new Color(87, 152, 64); // dark green, maybe good for contest?
		
		
		// Original color choices.
		//	Color flashCol = new Color(1.0f,.7f,0.0f); // orange
		//	Color foundCol = new Color(1.0f,0.0f,1.0f); // magenta
		//	Color diffCol = new Color(1.0f,.0f,.0f); //red
		//	Color markCol[] = new Color[4];
		//	markCol[0] = new Color(.0f,1.0f,.0f); // green
		//	markCol[1] = new Color(.0f,.0f,1.0f); // blue
		//	markCol[2] = new Color(.0f,1.0f,1.0f); // cyan
		//	markCol[3] = new Color(1.0f,1.0f,.0f); // yellow
		//	labelColor = new Color(.9f,.9f,.9f); // unused
		//	labelBackColor = new Color(0.0f,0.0f,0.0f); // black
		//	labelHiColor = new Color(0.0f,0.0f,0.0f); // black
		//	labelBackHiColor = new Color(1.0f,.7f,0.0f); // orange
		//	objectColor = new Color(.7f,.7f,.7f); // white
		//	backgroundColor = new Color(0.0f,0.0f,0.0f); // background black

		// the order of adding groups matters: the *first* group found
		// gets precedence for color. so the mouseover
		// highlight trumps everything, then the leaf selection.
		// finally, we want the structural differences to
		// take priority over the manual highlighting.

		for (int i = 0; i < 6; i++) {
			siteGroup[i] = addGroup(false, Sequence.getColor(i));
			setGroupEnable(siteGroup[i], false);
		}
		for (int i = 5; i >= 0; i--)
			setGroupPriority(siteGroup[i]);
		flashGroup = addGroup(false, flashCol);
		// enabled by default, unset since should be off
		setGroupEnable(flashGroup, false);
		diffGroupMark = addGroup(true, diffCol);
		// diff group should be marked across all sequences
		diffGroupDraw = addGroup(false, null);
		setGroupEnable(diffGroupDraw, false);
		gapGroup = addGroup(true, null); 
		System.out.println("gaps is group number: " + gapGroup);
		setGroupEnable(gapGroup, true);  // disable looking up colors on gaps, just a second
		// gap group should be collapsable across all sequences
		foundGroupDraw = addGroup(false, null);
		// found group has ranges with unique sequences
		foundGroupMark = addGroup(true, foundCol);
		
		rangeGapGroup = addGroup(true, null);
//		
		// found group has ranges with unique sequences
		setGroupEnable(foundGroupDraw, false);
		
		
		// the "draw" is for resizing, not marking
		setShowDiffs(true);


		quasimode = false;
		wantsFocusATD = null;

		edgeweightLevels = 1;
		edgeweightLevelUsed = 0;
		
		difference = 0.62f;

		foundFeatures = new Vector();
		
		// do UI after initialising all values
		ui = new UI(this, "SequenceJuxtaposer");

		mainFrame = ui.getMainFrame();
		//mainFrame.setSize(300, 300);
		//		drawPanel.setSize(300, 350);
		//mainFrame.setSize(width, height+controlPanel.getHeight());
		mainFrame.add(optionBar, BorderLayout.NORTH);
		drawPanel = ui.getDrawPanel();
	}

	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// this is now separated for annotations because you want to have this
	// fileindependent and dynamically
	
	public void addAnnotationGroups()
	{
		annotationGroupDraw = new int[maxAnnotationNum];
		annotationGroupMark = new int[maxAnnotationNum];
		for( int i = 0; i < maxAnnotationNum; i++)
		{
				annotationGroupMark[i] = addGroup(false, Annotation.getColor( ((String)foundFeatures.get(i)) ));
				annotationGroupDraw[i] = addGroup(false, null);
				setGroupEnable(annotationGroupDraw[i], false);
				Annotation.setGroupByName( ((String)foundFeatures.get(i)), annotationGroupMark[i] );
				//Annotation.setGroupByName( ((String)foundFeatures.get(i)), annotationGroupDraw[i] );
		}
		
	}

	/*
	void lostFocusInQuasi(AccordionTreeDrawer atd) {
	lostFocusATD = atd;
	}
	*/
	void wantsFocusInQuasi(AccordionSequenceDrawer atd) {
		wantsFocusATD = atd;
	}

	void setQuasimode(boolean on) {
		if (quasimode == true && on == false && wantsFocusATD != null) {
			//lostFocusATD.transferFocus();
			wantsFocusATD.requestFocus();
		}
		quasimode = on;
		//quasiATD = null;
		wantsFocusATD = null;
	}
	boolean getQuasimode() {
		return quasimode;
	}

	public void setTitle(String t) {
		title = t;
		mainFrame.setTitle(title);
	}

	/**
	 * Adds a new tree to "trees", performs all the pairwise tree
	 * comparisons and stores results.
	 *
	 * @author  Tamara Munzner
	 *
	 * @see AccordionDrawer.Tree
	 * @see TreePairs.addTree
	 */

	public void addSequence(Sequence newSequence) {
		if (newSequence.getNumOfNodes() > maxSeqLength)
			maxSeqLength = newSequence.getIterator().getNumOfChars();

		/**
		 * Construct the data structures for structural comparison and
		 * highlight
		 **/
		sequences.add(newSequence);
	}

	public Sequence getFirstSequence() {
		return getSequence(0);
	}

	public void addAnnotations( Vector list, Sequence seq )
		{
			//unmarkGroup(groupDraw);
			//unmarkGroup(groupMark);
		
			for( int i=0; i<list.size(); i++)
			{
				String name = ((Annotation)list.get(i)).getName();
				RangeList draw = getGroup( Annotation.getGroupByName( name ) );
				RangeList mark = getGroup( Annotation.getGroupByName( name )  );
			
				RangeList rl = new RangeList(-1);
				
				
				ArrayList ranges = ((Annotation)list.get(i)).getRanges();
				
				for( int j=0; j<ranges.size(); j++ )
				{
					int[] minmax = (int[])ranges.get(j);
					int min = minmax[0];
					int max = minmax[1];
				
					rl.addRange( min, max, seq, false, true );
							
					rl.setKey(annotationGroupDraw[i]);
					addNodesToGroup(rl, annotationGroupDraw[i], seq);
					rl.setKey(Annotation.getGroupByName( name ));
					addNodesToGroup(rl, Annotation.getGroupByName( name ));
			
				}
				
			
			
				RangeList groupMarkList = getGroup(Annotation.getGroupByName( name )); 
				if (rl == null || groupMarkList == null) {
					groupMarkList.clear();
				} else {
					//	done adding nodes to groupDraw and groupMark, recreate the rangeTrees for color range lookups
					groupMarkList.setColorTree(new ColorTree(groupMarkList));
					//System.out.println("made this rangeTree for found: " + groupMarkList.getRangeTree());
			}
		
			RangeList temp = getGroup(annotationGroupDraw[i]);
			temp.setColorTree(new ColorTree(temp));
			temp = getGroup(Annotation.getGroupByName( name ));
			temp.setColorTree(new ColorTree(temp));
			
			}	
		
			requestRedrawAll();
			//System.out.println("add annotations here");
		
		}

	public void createASD() {
			/**
			 * Create and add the drawer(s) for the tree
			 **/
			boolean gapAny = true;
			// true => a gap on any sitenode position will add the position to the gapgroup
			// false => only if all sitenodes in the position are gaps the sitenode position will be added to the gapgroup
			// TODO: implement false branch
			int height = 648;
			int width = 540;
			int first = -1, last = -1;
			boolean added = false;
			
			///////////////// REMOVE GAPS ///////////////////////////////////////////////////////////
			System.out.println("doing gap processing ... ");
			numSequences = sequences.size();
			for (int i = 0; i < numSequences; i++) {
					Sequence seqA = getSequence(i);

						seqA.setKey(i);
						if (seqA.getNumOfNodes() < maxSeqLength)
							seqA.lengthen(maxSeqLength);
					}

		
		
			int diffsSoFar = 0;
			//System.out.println("first of all: ");
			//getSequence(0).printAsPositions();
			
			
			
			
			int stringArraySize = getSequence(0).getIterator().arrayLength();
			for( int k=0; k< stringArraySize; k++)
			{
				int start = -2, end = -2;
//				int subStringSize = getSequence(0).getIterator().getSubStringByPosition(k).length();
//				int pos = k*subStringSize;

				for (int i = 0; i < getSequence(0).getIterator().getSubStringByPosition(k).length(); i++)
				{
						boolean foundGap = true;
						
						for (int j = 0; j < numSequences && foundGap; j++) { // we only care about whole columns of gaps
								//SiteNode n = 
								String str =  getSequence(j).getIterator().getSubStringByPosition(k);
								char n = str.charAt(i);
								//System.out.println("found node " + getSequence(0).getNumOfNodes() + " at position " + i + " char " + n.getName() + 
							// " " + getSequence(0).getNodeName(i+1) + " in sequence " + getSequence(j) + " seqString: " + getSequence(j).getString(0,getSequence(j).getNumOfNodes()-1));
								//if(str.length() != maxSeqLength){
									if (SiteNode.isGap(n)) {
									
										if (first == -1 && last == -1)
											first = last = j;
										else if (last == j - 1)
											last = j;
									
										foundGap = true;
									} else 
										foundGap = false;

								//}
									//System.out.println("pos j " + j + " of " + numSequences);
						}
						if(foundGap) // remove the gap
						{
							//
							if (end == -2)
								start = end = i;
							else if( i - 1 == end )
								end = i;
							else 
							{
								diffsSoFar = end - start+1;
								for(int b= 0; b<numSequences; b++)
								{
									Sequence s = getSequence(b);
									s.getIterator().removeAtPosition(k, start, diffsSoFar);
									
										//getSequence(b).removeSiteNode(i, diffsSoFar);
								}
								maxSeqLength -= diffsSoFar;
								i -= diffsSoFar;
								//System.out.println("size " + subStringSize);
								start = i;
								end = i;
							}
		
						}
						first = -1; 
						last = -1;
					
				}
				if (end != -2)
				{
					diffsSoFar = end - start + 1;
					for (int b = 0; b < numSequences; b++) {
						Sequence s = getSequence(b);
						s.getIterator().removeAtPosition(k, start, diffsSoFar);
						//getSequence(b).removeSiteNode(i, diffsSoFar);
					}
					maxSeqLength -= diffsSoFar;
				}
			}
			
			
			// rebuild the string arrays
			for (int b = 0; b < numSequences; b++)
			{
					getSequence(b).getIterator().rebuildArrayList();
					getSequence(b).createSites();									
			}

						
			
			/////////////// REMOVE GAPS  END///////////////////////////////////////////////////////////
		
			for (int i = 0; i < numSequences; i++) {
				Sequence seqA = getSequence(i);

				seqA.setKey(i);
				if (seqA.getNumOfNodes() > maxSeqLength)
					seqA.lengthen(maxSeqLength);
			}

			addASD(width, height);


			for (int i = 0; i < numSequences; i++) {
				Sequence seqA = getSequence(i);
				int namewidth =
					AccordionSequenceDrawer.stringWidth(seqA.getName(), 17);
				// 14 is the initial font size, be nice
				if (namewidth > Sequence.maxFontLength)
					Sequence.maxFontLength = namewidth;
				//System.out.println("MFL: " + Sequence.maxFontLength);
			}


			Sequence seq0 = getSequence(0);
			// calculate gaps
			RangeList gaps = getGroup(gapGroup);

		
		
			for (int i = 0; i < maxSeqLength; i++)
			{
				boolean foundGap = false;
				for (int j = 0; !foundGap && j < numSequences; j++) {
					SiteNode n = getSequence(j).getNode(i);
					if (n.isGap()) {
						added = false;
						if (first == -1 && last == -1)
							first = last = i;
						else if (last == i - 1)
							last = i;
						else {
							gaps.getRanges().add(new RangeInSequence(first, last, seq0));
							first = last = n.getKey(); // next range
							added = false;
						}
						foundGap = true;
					}
				}
			}
			if (added == false && first != -1)
				gaps.getRanges().add(new RangeInSequence(first, last, seq0));

//			System.out.println("gap ranges: " + gaps);
//			System.out.println("END OF GAP RANGES");
				
			// calculate diffs
		
		
		calculateDifferences(diffGroupDraw, diffGroupMark);
		
		//	because AD should not know anything about range here
		// except rangegap, we have to pass the groupsize and add this
		// one last to the group to not confuse the group numbers
		calculateRangeGaps( rangeGapGroup );

		
		
		System.out.println("adding annotations...");
		for(int i=0; i<sequences.size(); i++)
		{
			if( getSequence(i).getAnnotationList() != null )
				addAnnotations( getSequence(i).getAnnotationList(), getSequence(i) );
										
		}
		System.out.println("annotations added...");
							//					for (int i = 0; i < sequences.size(); i++)
				//					{
				//						Sequence seqA = (Sequence) sequences.get(i);
				//						if (seqA.getNumOfNodes() < maxSeqLength)
				//							seqA.lengthen(maxSeqLength);
				//						for (int j = i+1; j < sequences.size(); j++)
				//						{
				//							Sequence seqB = (Sequence) sequences.get(j);
				//							doStructuralDiff(seqA, seqB, (AccordionSequenceDrawer)treeDrawers.get(0), diffGroup, gapAny);
				//							//System.out.println("diffGroup: " + getGroup(diffGroup));
				//							//doStructuralDiff(seqB, seqA, (AccordionSequenceDrawer)treeDrawers.get(0), diffGroup); // only need to do it one way with 2d layout
				//						}
				//					}
			//}
			//				System.out.println("diffGroup Draw: " + getGroup(diffGroupDraw));
//						System.out.println("diffGroup Mark: " + getGroup(diffGroupMark));
//			System.out.println("END OF DIFFGROUP MARK RANGES");
			//				System.out.println("sequenceCount "+ numSequences);
			// Associate new range trees with the range lists
			
			
			
			Component c = asd.getCanvas();
			BorderLayout g = (BorderLayout) mainFrame.getLayout();

			drawPanel.add(c);
			searchPanel.setLayout(new BorderLayout());
			searchField = new TextField(30);
			searchField.addKeyListener(this);
			searchButton = new Button(searchButtonText);
			searchButton.addActionListener(this);
			searchResultsView = new TextField(15);
			searchResultsView.setEditable(false);
			mouseOverView = new TextField(15);
			mouseOverView.setEditable(false);
			Panel viewPanel = new Panel();
			viewPanel.setLayout(new GridLayout(1, 2, 0, 4));
			viewPanel.add(mouseOverView);
			viewPanel.add(searchResultsView);
			searchPanel.add(searchField, BorderLayout.CENTER);
			searchPanel.add(searchButton, BorderLayout.EAST);
			searchPanel.add(viewPanel, BorderLayout.WEST);

			mainFrame.add(drawPanel, BorderLayout.CENTER);
			mainFrame.add(searchPanel, BorderLayout.SOUTH);

			//mainFrame.pack();

			drawPanel.validate();
			//mainFrame.validate();
			//mainFrame.repaint();
			drawPanel.repaint();
			requestRedrawAll();

			ui.showGroupFrame();

			System.out.println("end of create asd");
		}

		void calculateRangeGaps( int group )
		{
			System.out.println(" doing rangeGap group precomputation ");
			for (int k = 0; k < numSequences; k++) // look in all sequences
			{
				
				for( int i=0; i<maxSeqLength; i++)
				{
						int i_start = i;
						char currLetter =getSequence(k).getNode(i).getName().charAt(0);
					
						while( currLetter == 'N' || currLetter == '-' )
						{
							i++;
							if( i == maxSeqLength ) break;
							currLetter =getSequence(k).getNode(i).getName().charAt(0);
							//System.out.println("seq " + i + " " + k);
						}
						if( i_start != i )
							addNodesToGroup(i_start, i, group, getSequence(k));				
				}
			}
		
			RangeList temp = getGroup(group);
			temp.setColorTree(new ColorTree(temp));
		
			//asd.rangeGapList = getGroup(group); 
			//asd.rangeGapList.setColorTree(new ColorTree(asd.rangeGapList));
		}

		public void addNodesToGroup(int min, int max, int group, Sequence seq, boolean checkCollide) {
			RangeList addToGroup = (RangeList) groupsByIndex.get(group);
			addToGroup.addRange(min, max, seq, group == foundGroupDraw, checkCollide);
			//System.out.println("range now has " + ((RangeList)groupsByIndex.get(group)).ranges.size() + " elements");
			//System.out.println("added min " + min + " max " + max + " group " + group + " tree " + atd.getKey());
		}

		public void calculateDifferences(int groupDraw, int groupMark)
			{
		
				unmarkGroup(groupDraw);
				unmarkGroup(groupMark);
			
				RangeList diffDraw = getGroup(groupDraw);
				RangeList diffMark = getGroup(groupMark);
				int firstGroupDraw = -1;
				int lastGroupDraw = -1;
				int[] lastGroupMark = new int[numSequences];
				int[] firstGroupMark = new int[numSequences];
				for (int i = 0; i < numSequences; i++)
					lastGroupMark[i] = firstGroupMark[i] = -1; // init
			
			
				if (showStructDiffs == true && numSequences > 1) {
				for (int i = 0; i < maxSeqLength; i++)
				{
						if (i % 1000 == 0)
							System.out.println("doing diffs, at position: " + i);
	
						int j = 0;
						char letter =getSequence(j).getNode(i).getName().charAt(0);
						while (j < numSequences
							&& (letter == 'N'
								|| letter == '-')) // find first non-wildcard
							{
							j++;
							if (j == numSequences) { break;} 
							letter =getSequence(j).getNode(i).getName().charAt(0);
						}
						if (j == numSequences) {
		//					System.out.println("column " + i + " is all spaces or all N's");
							continue; // next letter, only spaces or N's
						}
					
						boolean found = false;
						// show only differences higher than 10% 
						int counterA = 0;
						int counterC = 0;
						int counterG = 0;
						int counterT = 0;
						int counterU = 0;
					
						for (int k = j; !found && k < numSequences; k++) // look in all sequences
						{
							SiteNode currNode = getSequence(k).getNode(i);
							char currLetter = currNode.getName().charAt(0);
							if (currLetter == 'A')
								counterA++;
							else if (currLetter == 'C')
								counterC++;
							else if (currLetter == 'G')
								counterG++;
							else if (currLetter == 'T')
								counterT++;
							else if (currLetter == 'U')
								counterU++;
						}
					
						//		for biggest:
						char biggest;
						int maxAC = Math.max(counterA, counterC);
						if (counterA == maxAC)
							 biggest = 'A';
						else
							 biggest = 'C';
						if (counterT > maxAC)
							 biggest = 'T';
						int maxACT = Math.max(maxAC, counterT);
						if (counterG > maxACT)
							 biggest = 'G';
						int maxACGT = Math.max(maxACT, counterG);
						if (counterU > maxACGT)
							 biggest = 'U';
						int maxACGTU = Math.max(maxACGT, counterU);
						int numOfCountedSeqs = counterA + counterC + counterG + counterT + counterU;
						//						System.out.println(counterA + " " + counterC + " " + counterT + " " + counterG + " " + counterU);
						float allowedDiff = (numOfCountedSeqs * difference);
				
						if (numOfCountedSeqs == 0)
							continue;
						if ((numOfCountedSeqs - maxACGTU) < allowedDiff)
						{	
							if (lastGroupDraw != -1)
								addNodesToGroup(firstGroupDraw, lastGroupDraw, groupDraw, getSequence(0), false); // previous group
							firstGroupDraw = lastGroupDraw = -1;
							for (int k = 0; k < numSequences; k++) {
								if (lastGroupMark[k] != -1)
								{
									Sequence s = getSequence(k);
									addNodesToGroup(firstGroupMark[k], lastGroupMark[k], groupMark, s, false); // previous group
								}
								firstGroupMark[k] = lastGroupMark[k] = -1;
							}										
							continue;
						}
					
						// add some nodes to the diff group
						//diffDraw.setKey(groupDraw);
						if (firstGroupDraw == -1) firstGroupDraw = i; // first after init
						if (lastGroupDraw != i-1 && lastGroupDraw != -1)
						{ // add nodes to the drawing (resizing) group
							addNodesToGroup(firstGroupDraw, lastGroupDraw, groupDraw, getSequence(0), false); // previous group
							firstGroupDraw = i; // start new group
						}
						lastGroupDraw = i;
						// we're drawing, and adding to 1st sequence
						 found = true;
						 for (int k = j; k < numSequences; k++) {
								 Sequence s = getSequence(k);
								 //diffMark.setKey(groupMark);
								 char currLetter =s.getNode(i).getName().toUpperCase().charAt(0);
								 if (currLetter != biggest && currLetter != 'N' && currLetter != '-') {
									if (firstGroupMark[k] == -1) firstGroupMark[k] = i; // first after init
									if (lastGroupMark[k] != i-1 && lastGroupMark[k] != -1)
									{ // add nodes to the marking (red marks) group(s)
										addNodesToGroup(firstGroupMark[k], lastGroupMark[k], groupMark, s, false); // previous group
										firstGroupMark[k] = i; // start new group
									}
									lastGroupMark[k] = i;
								 }	
						 }
				}
					
						// no differences, next letter
				}
				if (lastGroupDraw == maxSeqLength-1) // need to add last column to diffs
				{
					addNodesToGroup(firstGroupDraw, lastGroupDraw, groupDraw, getSequence(0), false);
					for (int i = 0; i < numSequences; i++)
					{
						Sequence s = getSequence(i);
						if (lastGroupMark[i] == maxSeqLength-1 && lastGroupMark[i] != -1)
							addNodesToGroup(firstGroupMark[i], lastGroupMark[i], groupMark, s, false);
					}
				}
	
				
				RangeList temp = getGroup(groupDraw);
						temp.setColorTree(new ColorTree(temp));
						temp = getGroup(diffGroupMark);
						temp.setColorTree(new ColorTree(temp));
			
				requestRedrawAll();			
			}

	/**
	 * Utility function to make new AccordionTreeDrawer, set initial
	 * parameters and do treeDrawer list bookkeeping.
	 * @author Tamara Munzner
	 */
	private AccordionSequenceDrawerFinal addASD(int width, int height) {
		//AccordionSequenceDrawerFinal ATD = new AccordionSequenceDrawerFinal(newSeq, width, height, this);
		AccordionSequenceDrawerFinal ATD =
			new AccordionSequenceDrawerFinal(sequences, width, height, this);
		ATD.setBackgroundColor(backgroundColor);
		//ATD.setObjectColor(objectColor);
		ATD.setLabelColor(labelColor);
		ATD.setLabelBackColor(labelBackColor);
		ATD.setLabelHiColor(labelHiColor);
		ATD.setLabelBackHiColor(labelBackHiColor);
		ATD.setRubberbandColor(Color.getHSBColor(0.0f / 360f, .0f, 0.3f));
		ATD.setKey(0);
		ATD.linethickness = linethickness;
		ATD.setExpandLeaves(0);
		ATD.showdiffs = showStructDiffs;
		return ATD;
	}

	/**
	 * Delete a tree from "trees", clean up all the data structures
	 * constructed for tree comparisons. It is currently empty.
	 *
	 * @author  Li Zhang
	 *
	 * @see AccordionDrawer.Tree
	 * @see TreePairs.deleteTree
	 */
	public void deleteSequence(Sequence s) {

		System.out.println("Delete the tree " + s.getName() + ". To be done");
	}

	/**
	 * Fetch a tree by its name.  This is done by linear search as we
	 * do not expect to have many trees in memory.
	 *
	 * @author Li Zhang
	 */
	public Sequence getTreeByName(String name) {

		for (int i = 0; i < numSequences; i++) {
			Sequence t = (Sequence) sequences.get(i);
			if (t.getName().equals(name))
				return t;
		}
		return null;
	}

	protected void addAction() {
		ui.addAction();
	}
	
	protected void quitAction() {
		ui.quitAction();
	}

	public void changedMarks()
	{
		asd.changedMarks();
	}

	public void requestRedrawAll() {
		asd.requestRedraw();
	}

	public int addGroup(boolean thisTreeOnly, Color col) {
		int groupCount = groups.size();
		RangeList g = new RangeList(groupCount, col, true /* enabled */
		, thisTreeOnly);
		System.out.println("adding group " + groupCount);
		groups.add(g);
		groupsByIndex.add(g); // permanent ordering
		return groupCount;
	}

	public void setGroupPriority(int group) {
		Object g = groupsByIndex.get(group);
		if (group > diffGroupMark) { // don't touch first three
			groups.remove(g);
			groups.add(diffGroupMark + 1, g);
			// put it first of changeable ones
		}
	}

	// deliberately don't have removeGroup method. if you don't want
	// it anymore, disable it. why? we want the index in the groups
	// vector to match the key in the rangelist. maybe there's a less
	// hardline way to do this?...

	// what sort of error checking should we do here if group is out
	// of bounds? be silent, return null where appropriate? print out
	// error message? throw an informative exception? ...

	public void setGroupColor(int group, Color c) {
		((RangeList) groupsByIndex.get(group)).setColor(c);
	}

	public void setGroupEnable(int group, boolean on) {
		((RangeList) groupsByIndex.get(group)).setEnabled(on);
	}

	public void addNodesToGroup(int min, int max, int group, Sequence seq) {
		RangeList addToGroup = (RangeList) groupsByIndex.get(group);
		addToGroup.addRange(min, max, seq, false, group == foundGroupDraw);
		//System.out.println("range now has " + ((RangeList)groupsByIndex.get(group)).ranges.size() + " elements");
		//System.out.println("added min " + min + " max " + max + " group " + group + " tree " + atd.getKey());
	}

	public void addNodesToGroup(RangeList r, int group, Sequence s0) {
		Iterator iter = r.getRanges().iterator();
		//System.out.println("adding range " + r + " to sequence 0 for group " + group + " which had " + getGroup(group));
		while (iter.hasNext()) {
			RangeInSequence currRange = (RangeInSequence) iter.next();
			addNodesToGroup(currRange.getMin(), currRange.getMax(), group, s0);
		}
	}

	public void addNodesToGroup(RangeList r, int group) {
		Iterator iter = r.getRanges().iterator();
		//System.out.println("adding range " + r + " to sequences for group " + group + " which had " + getGroup(group));
		while (iter.hasNext()) {
			RangeInSequence currRange = (RangeInSequence) iter.next();
			addNodesToGroup(
				currRange.getMin(),
				currRange.getMax(),
				group,
				currRange.getSeq());
		}
	}

	public void removeNodesFromGroup(
		int min,
		int max,
		int group,
		Sequence seq) {
		((RangeList) groupsByIndex.get(group)).removeRange(min, max, seq);
	}

	public Color getGroupColor(int group) {
		return ((RangeList) groupsByIndex.get(group)).getColor();
	}

	public void clearGroup(int group) {
		((RangeList) groupsByIndex.get(group)).clear();
	}

	public void unmarkGroup(int group) {
		changedMarks();
		clearGroup(group);
	}

	public RangeList getGroup(int group) {
		return ((RangeList) groupsByIndex.get(group));
	}

	// seed paths all the way up to the root
	void seedQueue(AccordionSequenceDrawer asd, boolean drawSpecial) {

		asd.clearQueue();
	
		Sequence seq0 = getSequence(0);
		//int stepSize = maxSeqLength/1000;
		int stepSize = maxSeqLength/sequences.size();
		if (drawSpecial)
		{
			int i;
			// draw default focus range first
			
					
			if( asd.defaultFocusRange[1] == 0 ) asd.defaultFocusRange[1] = maxSeqLength;
			for (i = asd.defaultFocusRange[0]-1;  i+stepSize+10 < asd.defaultFocusRange[1]; i+=stepSize)
			{
				asd.addRangeToDrawQueue(new RangeInSequence(i, i+stepSize, seq0));
			}
			asd.addRangeToDrawQueue(new RangeInSequence(i, asd.defaultFocusRange[1]-1, seq0));
					
			// then sites around
			if( asd.defaultFocusRange[1] != maxSeqLength )
			{
				// from end of box til maxlength
				for (i = asd.defaultFocusRange[1]-1;  i+stepSize+10 < maxSeqLength; i+=stepSize)
				{
						asd.addRangeToDrawQueue(new RangeInSequence(i, i+stepSize, seq0));
				}
				asd.addRangeToDrawQueue(new RangeInSequence(i, maxSeqLength-1, seq0));
				// from begin of box til 0
				for (i = asd.defaultFocusRange[0];  i-stepSize-10 > 0; i-=stepSize)
				{
						asd.addRangeToDrawQueue(new RangeInSequence(i-stepSize,i , seq0));
				}
				asd.addRangeToDrawQueue(new RangeInSequence(-1, i, seq0));
			}
			// if we only select the last column
			if( asd.defaultFocusRange[1] == maxSeqLength && asd.defaultFocusRange[0] == maxSeqLength-1 ) 
			{
					// from begin of box til 0
					for (i = asd.defaultFocusRange[0];  i-stepSize-10 > 0; i-=stepSize)
					{
							asd.addRangeToDrawQueue(new RangeInSequence(i-stepSize,i , seq0));
					}
					asd.addRangeToDrawQueue(new RangeInSequence(-1, i, seq0));
			}
			
			// draw from 0 to end 
			// seed ranges of columns to draw
//			for (i = -1; i+stepSize+10 < maxSeqLength; i+=stepSize)
//			{
//				asd.addRangeToDrawQueue(new RangeInSequence(i, i+stepSize-1, seq0));
//			}
//			asd.addRangeToDrawQueue(new RangeInSequence(i, maxSeqLength-1, seq0));
//			
		}
		else
		{
		// seed with rootCell of every bitree
//			for (int i = 0; i < numSequences; i++) {
////			asd.biTreeRoots[i].computePlaceThisFrame(); // called here so comparator doesn't have to call it
//				asd.addCellToQueue(asd.biTreeRoots[i]);
//			//System.out.println("seeded "+asd.biTreeRoots[i]);
//			}
		}
	}


	// built-in group precedence ordering determines which group color
	// wins out. just stop when you find the first one, because order
	// matters.

	ArrayList getColorsForRange(
		int objMin,
		int objMax,
		int seqMin,
		int seqMax,
		AccordionSequenceDrawer callingSequenceDrawer) {
		RangeInSequence fakeRange = new RangeInSequence(objMin, objMax, getSequence(seqMin));
		Sequence seq0 = callingSequenceDrawer.getSequence(0);
		ArrayList returnList = new ArrayList(2);
		//	System.out.println("TOP objmin " + objMin + " objmax " + objMax + " seqmin " + seqMin + " seqmax " + seqMax );// + " callingtree " + callingTree.getKey());
		RangeList foundGroup = null;
		Iterator GroupIter = groups.iterator();
//		System.out.println("iterating through groups: " +  groups.size());
		while (GroupIter.hasNext()) {
			RangeList group = (RangeList) GroupIter.next();
			ColorTree tree = group.getColorTree();
			
			boolean thisTreeOnly = group.getThisTreeOnly();
			if (true == group.getEnabled() && tree != null)
			{
				RangeInSequence foundRange = tree.findSplitRangeNode(tree.getRoot(), fakeRange);
				if (foundRange != null && group.getColor() != null)
				{
					returnList.add(group.getColor());
					
				}
			}
		}
		
		return returnList;
	}

	/** 
	 * Acts as a toggle on group members: turn off old ones, turn on
	 * new ones, do lightweight frontbuffer drawing per item not whole
	 * screen redraw
	 *
	 * @author  Tamara Munzner
	 * @see AccordionDrawer.Tree
	 * @see AccordionDrawer.TreeNode
	 * @see TreeJuxtaposer.RangeList
	 *
	 */

	void doFlashGeom(
		CellGeom flashGeom,
		int group,
		AccordionSequenceDrawerFinal thisSequence,
		int x,
		int y,
		boolean all) {
		//System.out.println("doFlashGeom: " + x + " " + y);
		RangeList theGroup = getGroup(group);
		SiteNode flashNode;
		//System.out.println("fg "+flashGeom+" fgo "+flashGeomOld);
		if (null == flashGeom) {
			//	    if (null != flashGeomOld) {

			if (true) {
				// the first time flashGeom goes null, we need to call
				// setFlash with a null geometry to make sure the
				// highlight is turned off.
				flashNode = null;
			} else
				return;
		} else {
			// flashGeom is an edge. get its associated node instead.
			//	    flashNode = thisSequence.getNode(flashGeom.getKey());
			flashNode = (SiteNode) flashGeom;

							//System.out.println("flash normal");
		}
		//System.out.println("fn "+flashNode+" fgo "+flashGeomOld);
		if (flashNode == flashGeomOld)
		{
			System.out.println("flash same");
			return;
		}
		// we're still doing the group book-keeping so that we know
		// which tree the item is attached to. we just don't want to
		// enable the group, since then we're left with yellow boxes
		// that aren't easy to clean up with this frontbuffer trick!
		clearGroup(group);
		if (null == flashNode)
			return;

		addNodesToGroup(flashNode.getKey(), flashNode.getKey(), group, flashNode.getSequence());
		Sequence flashSeq = flashNode.getSequence();
		asd.setFlash((CellGeom) flashNode, theGroup.getColor(), x, y, true);

		//System.out.println("TJ set fgo "+flashNode);
		flashGeomOld = flashNode;
	}

	void doSelectGeoms(ArrayList attachedGeoms, boolean selectSubtree,
		int group, AccordionSequenceDrawer thisSequence)
	{
		clearGroup(group);
		if (null != attachedGeoms)
		{
			Iterator geomIter = attachedGeoms.iterator();
			while (geomIter.hasNext())
			{
				CellGeom cg = (CellGeom)geomIter.next();
				doSelectGeom(cg, selectSubtree, group, thisSequence);
				System.out.println("this tree node key " + cg.getKey() +
					" min "	+ cg.getMin() + " max "	+ cg.getMax() + 
					" thisSequence " + thisSequence.getKey());
			}
		}
		requestRedrawAll();
	}

	void doSelectGeom(CellGeom g, boolean selectSubtree, int group,
		AccordionSequenceDrawer thisSequence)
	{
		System.out.println("key to select: " + g.getKey());
		int which = (true == selectSubtree) ? g.getMax() : g.getKey();
		addNodesToGroup(g.getKey(), which, group, ((SiteNode) g).getSequence());
		changedMarks();
	}

	void setFocus(CellGeom g, AccordionSequenceDrawer atd) {
		atd.setFocusCell(g.getCell());
	}

	public void setBigGrid(boolean on) {

		asd.setBigGrid(on);
	}
	public void toggleBigGrid() {
		asd.toggleBigGrid();
	}

	public void setCacheRange(boolean on) {
		asd.setCacheRange(on);
	}
	public void toggleCacheRange() {
		asd.toggleCacheRange();
	}
	public void setCheckTime(boolean on) {
		asd.setCheckTime(on);
	}
	public void toggleCheckTime() {
		asd.toggleCheckTime();
	}
	public void setColorGrid(boolean on) {
		asd.setColorGrid(on);
	}
	public void toggleColorGrid() {
		asd.toggleColorGrid();
	}

	public void setDimBrite(boolean on) {
		asd.setDimBrite(on);
	}
	public void toggleDimBrite() {
		asd.toggleDimBrite();
	}
	public boolean getDimBrite() {
			return asd.dimbrite;
	}

	public void setDimColors(boolean on) {
		asd.setDimColors(on);
	}
	public void toggleDimColors() {
		asd.toggleDimColors();
	}
	public boolean getDimColors() {
		return asd.dimcolors;

	}

	public void setDrawLabels(boolean on) {
		asd.setDrawLabels(on);
	}
	public void toggleDrawLabels() {
		asd.toggleDrawLabels();
	}
	public boolean getDrawLabels() {
			return asd.drawlabels;
	}

	public void setDrawOrder(int order) {
		asd.setDrawOrder(order);
	}
	public void toggleDrawOrder() {
		asd.toggleDrawOrder();
	}

	public void setDumpStats(boolean on) {
		asd.setDumpStats(on);
	}
	public void toggleDumpStats() {
		asd.toggleDumpStats();
	}
	public void setGuarVis(boolean on) {
		asd.setGuarVis(on);
	}
	public void toggleGuarVis() {
		asd.toggleGuarVis();
	}

	public void setExpandLeaves(int expand) {
		asd.setExpandLeaves(expand);
	}
	public void toggleExpandLeaves() {
		asd.toggleExpandLeaves();
	}

	public void setLabelBuffer(int buffer) {
		asd.setLabelBuffer(buffer);
	}
	public void increaseLabelBuffer() {
		asd.increaseLabelBuffer();
	}
	public void decreaseLabelBuffer() {
		asd.decreaseLabelBuffer();
	}
	public void setLabelBuffer(int buffer, int xy) {
		asd.setLabelBuffer(buffer, xy);
	}
	public void increaseLabelBuffer(int xy) {
		asd.increaseLabelBuffer(xy);
	}
	public void decreaseLabelBuffer(int xy) {
		asd.decreaseLabelBuffer(xy);
	}

	public void setLabelDrawBack(boolean on) {
		asd.setLabelDrawBack(on);
	}
	public void toggleLabelDrawBack() {
		asd.toggleLabelDrawBack();
	}

	public void setLabelPopup(boolean on) {
		asd.setLabelPopup(on);
	}
	public void toggleLabelPopup() {
		asd.toggleLabelPopup();
	}

	public void setLabelPosRight(boolean on) {
		asd.setLabelPosRight(on);
	}
	public void toggleLabelPosRight() {
		asd.toggleLabelPosRight();
	}

	public void setLabelTransp(boolean on) {
		asd.setLabelTransp(on);
	}
	public void toggleLabelTransp() {
		asd.toggleLabelTransp();
	}

	public void setLineThickness(float thickness) {
		asd.setLineThickness(thickness);
	}
	public void increaseLineThickness() {
		asd.increaseLineThickness();
	}
	public void decreaseLineThickness() {
		asd.decreaseLineThickness();
	}

	public void setLinkedNav(boolean on) {
		asd.setLinkedNav(on);
	}
	public void toggleLinkedNav() {
		asd.toggleLinkedNav();
	}
	public boolean getLinkedNav() {
		return asd.linkednav;
	}

	public void setMaxFontHeight(int fontsize) {
		asd.setMaxFontHeight(fontsize);
	}
	public void increaseMaxFontHeight() {
		asd.increaseMaxFontHeight();
	}
	public void decreaseMaxFontHeight() {
		asd.decreaseMaxFontHeight();
	}

	public void setMinFontHeight(int fontsize) {
		asd.setMinFontHeight(fontsize);
	}
	public void increaseMinFontHeight() {
		asd.increaseMinFontHeight();
	}
	public void decreaseMinFontHeight() {
		asd.decreaseMinFontHeight();
	}

	public void setNoFlash(boolean on) {
		asd.setNoFlash(on);
	}
	public void toggleNoFlash() {
		asd.toggleNoFlash();
	}

	public void setNoGeoms(boolean on) {
		asd.setNoGeoms(on);
	}
	public void toggleNoGeoms() {
		asd.toggleNoGeoms();
	}

	public void setNoGrid(boolean on) {
		asd.setNoGrid(on);
	}
	public void toggleNoGrid() {
		asd.toggleNoGrid();
	}

	public void setShowAlg(boolean on) {
		asd.setShowAlg(on);
	}
	
	public void drawAnnotations(String name, boolean flag)
	{
		for( int i = 0; i<maxAnnotationNum; i++)
		{
			if( name.equalsIgnoreCase( ((String)foundFeatures.get(i))))
				setGroupEnable( annotationGroupMark[i], flag);
				reset();
		}	
	}
		
	
	public void toggleShowAlg() {
		asd.toggleShowAlg();
	}

	public void setWildOn(boolean on) {
		boolean oldValue = nMatchesEverything;
		nMatchesEverything = on;
		if (oldValue != on)
			searchSequences(SequenceJuxtaposer.foundGroupDraw, SequenceJuxtaposer.foundGroupMark);
	}
	
	public void setDifferenceValue( float difference)
	{
		this.difference = difference;
		System.out.println("diff " + difference);
		calculateDifferences(diffGroupDraw, diffGroupMark);
		//changedMarks();
		//requestRedrawAll();
	}
	
	public float getDifferenceValue()
	{
		return this.difference;	
	}

	public void setShowDiffs(boolean on) {
		showStructDiffs = on;
		System.out.println("showdiffs " + on);
		setGroupEnable(diffGroupMark, on);
		if (asd != null)
		{
			asd.showdiffs = on;
			changedMarks();
			requestRedrawAll();
		}
	}
	public void toggleShowDiffs() {
		showStructDiffs = !showStructDiffs;
		setShowDiffs(showStructDiffs);
		//	setGroupEnable(diffGroup, showStructDiffs);
		//	changedMarks();
		//	requestRedrawAll();
	}
	public boolean getShowDiffs() {
		return showStructDiffs;
	}
	/*	public void setLCAGroup(boolean on) { stateFrame.setLCAGroup(on); }
		public boolean getLCAGroup() { return stateFrame.getLCAGroup(); }
		public void setLCAMark(boolean on) { stateFrame.setLCAMark(on); }
		public boolean getLCAMark() { return stateFrame.getLCAMark(); }
	*/

	public void reset() {
			asd.reset();
			//asd.drawSiteNodeDensity(asd.gridDepth[0] > asd.gridDepth[1] ? asd.gridDepth[0] : asd.gridDepth[1]);
	}

	/** Return one SiteNode per group range item.
	 * @author = Tamara Munzner, Serdar Tasiran
	 */
	public ArrayList getGroupForest(
		int groupindex,
		AccordionSequenceDrawer asd) {

		RangeList group = getGroup(groupindex);
		ArrayList forestRoots = new ArrayList();
		Iterator groupIter = group.getRanges().iterator();
		while (groupIter.hasNext()) {
			RangeInSequence r = (RangeInSequence) groupIter.next();
			//		if (r.getAsd() == asd) // always true with a 2d array of nodes
			{
				for (int i = r.getMin(); i <= r.getMax(); i++)
					forestRoots.add(r.getSeq().getNode(i));
			}
		}
		return forestRoots;
	}

	public void resizeGroup(
		int groupindex,
		boolean grow,
		int numSteps) { //, boolean horizontal, boolean vertical) {
		RangeList group = getGroup(groupindex);
		while (asd.transitionAnimating)
			asd.endAllTransitions(); // end transitions, new queue is going to be swapped in
		Hashtable newToMove = new Hashtable();
		if (!grow)
			group = (RangeList)group.flipRangeToShrink(AccordionDrawer.X, asd);
		asd.getSplitLine(AccordionDrawer.X).resizeForest(group, numSteps, newToMove, asd.getInflateIncr());
		asd.toMove = newToMove; // swap in new queue
//				System.out.println("Move queue: " + ad.toMove);
		asd.requestRedraw(); // redraw the frame, this will enact the movements in the move queue
	}

	public void loadFasta(String fileName, int maxLoad) {
		try {
			int lastDot = fileName.lastIndexOf(".");
			String extension = fileName.substring(lastDot+1);
			if( (extension.equalsIgnoreCase("fa")) ||
				(extension.equalsIgnoreCase("fasta")) ){
				SequenceLoader sl = new SequenceLoader(this, fileName, maxLoad);
			} else {
				FlatFileLoader sl = new FlatFileLoader(this, fileName, maxLoad);
			}
		} catch (FileNotFoundException e) {
			System.out.println("File not found " + fileName);
			e.printStackTrace();
		} 
//		mainFrame.setSize(500, 500);
//		drawPanel.setSize(500, 520);
		mainFrame.setVisible(true);

	}

	public static void main(String[] args) {

		SequenceJuxtaposer sj = new SequenceJuxtaposer();
		
		// get rid of annoying sync warnings
		System.setProperty("java.util.prefs.syncInterval", "2000000");
		int maxLoad = Integer.MAX_VALUE;
		int i = 0;
		while (i < args.length) {
			if (args[i].equals("-f")) // flag to fully qualify node names
				AccordionSequenceDrawer.fullyQualified = true;
			// default is false
			else if (args[i].equals("-nostructdiff")) {
				sj.setShowDiffs(false);
			} else if (args[i].equals("-n")) {
				maxLoad = Integer.valueOf(args[++i]).intValue();
			} else if (args[i].equals("-matrix")) {
				sj.matrix = true;
			} else if (args[i].equals("-L")) {
				sj.edgeweightLevels = (new Integer(args[++i])).intValue();
				sj.edgeweightLevelUsed = (new Integer(args[++i])).intValue();
				if (sj.edgeweightLevelUsed >= sj.edgeweightLevels)
					sj.edgeweightLevelUsed = sj.edgeweightLevels - 1;
				else if (sj.edgeweightLevelUsed < 0)
					sj.edgeweightLevelUsed = 0;
			} else if (args[i].equals("-x")) {
				i++;
				String filename = args[i++];

				/*Vector numbers = new Vector();
				while(i < args.length){
				    String s = args[i];
				    Integer n;
				    try{
					n = Integer.valueOf(s);
				    }catch(NumberFormatException ne){
					break;
				    }
				    numbers.add(n);
				    i++;
				}
				//sj.loadNexusTree(filename, numbers);
				*/
				// to offset the next operation i++
				i--;

			} else {
				sj.loadFasta(args[i], maxLoad);
				//sj.loadNewickTree(args[i]);
				sj.sequenceArgs = args.length - i;
			}
			i++;
		}

		System.out.println(
			"levels: "
				+ sj.edgeweightLevels
				+ " level used: "
				+ sj.edgeweightLevelUsed);
		//if (i == 0)
		//	sj.mainFrame.setSize(663, 500);
		sj.mainFrame.setSize(648, 550);
//		sj.drawPanel.setSize(648, 540);
		//sj.mainFrame.pack();
		//System.out.println("setting visble");
		sj.mainFrame.setVisible(true);
		//	System.out.println("finished setting visble");

		//    System.out.println("end of main");

	}

	public void observe() {
		//System.out.println("observing...");
		if (asd == null)
			return;
		observeActionGroup();
		observeActionMode();
		observeActionTarget();
		observeMarkGroup();
		//	stateFrame.observe();
	}

	// ugly implementation
	/** 
		* a more compact version of implementation for observing functions 
		* @author  Jeffrey Zhang
		*
		*/
	public void observeActionMode() {
		int treeDrawerActionMode = asd.getActionMode();
		//System.out.println("observeActionMode: " + treeDrawerActionMode);
		/*			if (treeDrawerActionMode == AccordionTreeDrawerFinal.MOUSEOVER)
						debugFrame.result[DebugFrame.ACT_MODE].setText(DebugFrame.actionModeLabel[treeDrawerActionMode]);
					else if (treeDrawerActionMode == AccordionTreeDrawerFinal.ST_FREEMOVE)
						debugFrame.result[DebugFrame.ACT_MODE].setText(DebugFrame.actionModeLabel[treeDrawerActionMode]);
					else if (treeDrawerActionMode == AccordionTreeDrawerFinal.ST_FREEMOVEAGAIN)
						debugFrame.result[DebugFrame.ACT_MODE].setText(DebugFrame.actionModeLabel[treeDrawerActionMode]);
					else if (treeDrawerActionMode == AccordionTreeDrawerFinal.ST_RESHAPE)
						debugFrame.result[DebugFrame.ACT_MODE].setText(DebugFrame.actionModeLabel[treeDrawerActionMode]);
					else if (treeDrawerActionMode == AccordionTreeDrawerFinal.RECT_CREATE)
						debugFrame.result[DebugFrame.ACT_MODE].setText(DebugFrame.actionModeLabel[treeDrawerActionMode]);
					else if (treeDrawerActionMode == AccordionTreeDrawerFinal.RECT_FREEMOVE)
						debugFrame.result[DebugFrame.ACT_MODE].setText(DebugFrame.actionModeLabel[treeDrawerActionMode]);
					else if (treeDrawerActionMode == AccordionTreeDrawerFinal.RECT_FREEMOVEAGAIN)
						debugFrame.result[DebugFrame.ACT_MODE].setText(DebugFrame.actionModeLabel[treeDrawerActionMode]);
					else if (treeDrawerActionMode == AccordionTreeDrawerFinal.RECT_RESHAPE)
						debugFrame.result[DebugFrame.ACT_MODE].setText(DebugFrame.actionModeLabel[treeDrawerActionMode]);
					else
						debugFrame.result[DebugFrame.ACT_MODE].setText("unknown action mode: " + treeDrawerActionMode);
			*/
		//debugFrame.result[DebugFrame.ACT_MODE].setText(DebugFrame.actionModeLabel[treeDrawerActionMode]);

	}

	public void observeActionGroup() {
		int treeDrawerActionGroup = asd.getActionGroup();
		//System.out.println("observeActionGroup: " + treeDrawerActionGroup);
		/*			if (treeDrawerActionGroup == TreeJuxtaposer.markGroup[0])
						debugFrame.result[DebugFrame.ACT_TYPE].setText(StateFrame.activeLabel[StateFrame.GA_ACT]);
					else if (treeDrawerActionGroup == TreeJuxtaposer.markGroup[1])
						debugFrame.result[DebugFrame.ACT_TYPE].setText(StateFrame.activeLabel[StateFrame.GB_ACT]);
					else if (treeDrawerActionGroup == TreeJuxtaposer.markGroup[2])
						debugFrame.result[DebugFrame.ACT_TYPE].setText(StateFrame.activeLabel[StateFrame.GC_ACT]);
					else if (treeDrawerActionGroup == TreeJuxtaposer.markGroup[3])
						debugFrame.result[DebugFrame.ACT_TYPE].setText(StateFrame.activeLabel[StateFrame.GD_ACT]);
					else if (treeDrawerActionGroup == TreeJuxtaposer.flashGroup)
						debugFrame.result[DebugFrame.ACT_TYPE].setText(StateFrame.activeLabel[StateFrame.M_ACT]);
					else if (treeDrawerActionGroup == TreeJuxtaposer.diffGroup)
						debugFrame.result[DebugFrame.ACT_TYPE].setText(StateFrame.activeLabel[StateFrame.D_ACT]);
					else if (treeDrawerActionGroup == TreeJuxtaposer.foundGroup)
						debugFrame.result[DebugFrame.ACT_TYPE].setText(StateFrame.activeLabel[StateFrame.F_ACT]);
					else
						debugFrame.result[DebugFrame.ACT_TYPE].setText("unknown action group: " + treeDrawerActionGroup);
		*/

		//        debugFrame.result[DebugFrame.ACT_TYPE].setText(StateFrame.activeLabel[treeDrawerActionGroup]);
	}

	public void observeActionTarget() {
		// action target split into: selectionResolution and growDirection
		int treeDrawerSelectionResolution = asd.getSelectionResolution();
		int treeDrawerGrowDirection = asd.getGrowDirection();
		//System.out.println("observeActionTarget: " + treeDrawerActionTarget);
		/*			if (treeDrawerGrowDirection == AccordionTreeDrawerFinal.HORIZ)
						debugFrame.result[DebugFrame.ACT_GROWDIR].setText(DebugFrame.actionTargetLabel[DebugFrame.H_TARGET]);
					else if (treeDrawerGrowDirection == AccordionTreeDrawerFinal.VERT)
						debugFrame.result[DebugFrame.ACT_GROWDIR].setText(DebugFrame.actionTargetLabel[DebugFrame.V_TARGET]);
					else if (treeDrawerGrowDirection == AccordionTreeDrawerFinal.ALL)
						debugFrame.result[DebugFrame.ACT_GROWDIR].setText(DebugFrame.actionTargetLabel[DebugFrame.B_TARGET]);
					else
						debugFrame.result[DebugFrame.ACT_GROWDIR].setText("unknown grow direction: " + treeDrawerGrowDirection);
		*/
		//	debugFrame.result[DebugFrame.ACT_GROWDIR].setText(DebugFrame.actionTargetLabel[treeDrawerGrowDirection]);			

		/*			if (treeDrawerSelectionResolution  == AccordionTreeDrawerFinal.NODE)
						debugFrame.result[DebugFrame.ACT_SELECTION].setText(DebugFrame.actionTargetLabel[DebugFrame.NODE_TARGET]);
					else if (treeDrawerSelectionResolution == AccordionTreeDrawerFinal.SUBTREE)
						debugFrame.result[DebugFrame.ACT_SELECTION].setText(DebugFrame.actionTargetLabel[DebugFrame.SUBTREE_TARGET]);
					else
						debugFrame.result[DebugFrame.ACT_SELECTION].setText("unknown selection resolution: " + treeDrawerSelectionResolution );
		*/
		//   debugFrame.result[DebugFrame.ACT_SELECTION].setText(DebugFrame.actionTargetLabel[treeDrawerSelectionResolution + 3]);
	}

	public void observeMarkGroup() {
		int treeDrawerMarkGroup = asd.getMarkGroup();
		//System.out.println("observeMarkGroup: " + treeDrawerMarkGroup);
		/*		if (treeDrawerMarkGroup == markGroup[0])
					debugFrame.result[DebugFrame.M_GROUP].setText(StateFrame.activeLabel[StateFrame.GA_ACT]);
				else if (treeDrawerMarkGroup == markGroup[1])
					debugFrame.result[DebugFrame.M_GROUP].setText(StateFrame.activeLabel[StateFrame.GB_ACT]);
				else if (treeDrawerMarkGroup == markGroup[2])
					debugFrame.result[DebugFrame.M_GROUP].setText(StateFrame.activeLabel[StateFrame.GC_ACT]);
				else if (treeDrawerMarkGroup == markGroup[3])
					debugFrame.result[DebugFrame.M_GROUP].setText(StateFrame.activeLabel[StateFrame.GD_ACT]);
				else
					debugFrame.result[DebugFrame.M_GROUP].setText("unknown mark group: " + treeDrawerMarkGroup);
		*/
		//   debugFrame.result[DebugFrame.M_GROUP].setText(StateFrame.activeLabel[treeDrawerMarkGroup]);

	}

	public void windowClosing(WindowEvent event) {
		System.exit(0);
	}
	public void windowClosed(WindowEvent event) {
		System.exit(0);
	}
	public void windowActivated(WindowEvent event) {
		//System.out.println("window activated");
	}
	public void windowDeiconified(WindowEvent event) {
		//System.out.println("window deiconified");
		requestRedrawAll();
	}
	public void windowDeactivated(WindowEvent event) {
	}
	public void windowIconified(WindowEvent event) {
	}
	public void windowOpened(WindowEvent event) {
		requestRedrawAll();
	}

	public void searchSequences(int groupDraw, int groupMark) {
		//System.out.println("searching for ("+searchString.length()+"): " + searchString);
		final int cutoff = 100;
		// don't return more than this many search results
		if (searchString != null && searchString.length() > searchStringCutoffLength)
			searchField.setText(
				searchString.substring(0, searchStringCutoffSide)
					+ "..."
					+ searchString.substring(
						searchString.length() - searchStringCutoffSide));
		else
			searchField.setText(searchString);
		Iterator iter = sequences.iterator();
		unmarkGroup(groupDraw);
		unmarkGroup(groupMark);
		Sequence.maxFound = cutoff * numSequences;
		Sequence seq0 = getFirstSequence();
		RangeList r = null;
//		boolean nMatchesEverything = false;
		boolean protein = true;
		while (iter.hasNext()) {
			Sequence currSeq = (Sequence) iter.next();
			r = currSeq.search(searchString, cutoff, true, nMatchesEverything, protein);
			if (r == null) {
				//System.out.println("range is null: " + searchString);
				break;
			}
			Sequence.maxFound -= r.getNumRanges();
			r.setKey(groupDraw);
			addNodesToGroup(r, groupDraw, seq0);
			r = currSeq.search(searchString, cutoff, false, nMatchesEverything, protein);
			r.setKey(groupMark);
			addNodesToGroup(r, groupMark);
		}
		RangeList groupMarkList = getGroup(groupMark); 
		if (r == null || groupMarkList == null || Sequence.maxFound <= 0) {
			searchResultsView.setText(
				cutoff + "+ ranges in a sequence"); // limit is really cutoff*numOfSequences, shorter to write this
				// it's actually ok to have more than cutoff ranges in a sequence as long as the total is smaller than the real limit
			groupMarkList.clear();
		} else {
			searchResultsView.setText(
				searchString == null ? "" : searchString.length()
					+ " pairs long, "
					+ (cutoff * numSequences - Sequence.maxFound)
					+ " found");
		//	done adding nodes to groupDraw and groupMark, recreate the rangeTrees for color range lookups
			groupMarkList.setColorTree(new ColorTree(groupMarkList));
			//System.out.println("made this rangeTree for found: " + groupMarkList.getRangeTree());
		}
		requestRedrawAll();
		//		System.out.println("Draw total: " + groups.get(groupDraw));
		//		System.out.println("Mark total: " + groups.get(groupMark));
	}

	//	public void searchSequences()
	//	{
	//		searchSequences(searchField.getText(), foundGroupDraw, foundGroupMark);
	//	}

	public void grokTextField() {
		String fieldText = searchField.getText();
		int dotdotdotpos = fieldText.indexOf("...");
		if (dotdotdotpos >= 0) // text has been shortened
			{
			String start = fieldText.substring(0, dotdotdotpos);
			String end = fieldText.substring(dotdotdotpos + 3);
			//System.out.println("Start: " + start);
			//System.out.println("End: " + end);
			searchString =
				start
					+ searchString.substring(
						searchStringCutoffSide,
						searchString.length() - searchStringCutoffSide)
					+ end;
		} else
			searchString = fieldText.toUpperCase();
	}

	/* (non-Javadoc)
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == searchButton) {
			grokTextField();
			searchSequences(foundGroupDraw, foundGroupMark);
			requestRedrawAll();
		}

	}

	/* (non-Javadoc)
	 * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
	 */
	public void keyPressed(KeyEvent e) {
		int key = e.getKeyCode();
		if (key == KeyEvent.VK_ENTER) {
			//System.out.println("entered: " + searchString);
			grokTextField();
			searchSequences(foundGroupDraw, foundGroupMark);
			requestRedrawAll();
		}
	}

	/* (non-Javadoc)
	 * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
	 */
	public void keyReleased(KeyEvent e) {
	}

	/* (non-Javadoc)
	 * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
	 */
	public void keyTyped(KeyEvent e) {
	}

	public Sequence getSequence(int pos) {
		return (Sequence) sequences.get(pos);
	}
	
	

};
